/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "rdfskillcycles.h"

#include <QSet>
#include <QStack>

#include "../rdfsupdate.h"
#include "../rdfscursor.h"
#include "../rdfslogger.h"

using namespace Rdfs;

bool RdfsKillCycles::exec(const QStringList& arguments)
{
    QString predicate = arguments.at(0);
    RdfsTable table = RDFS_ORIGINAL;

    if (arguments.count() > 1 && arguments.at(1).compare("original", Qt::CaseInsensitive) == 0) {
        table = RDFS_ORIGINAL;
    }
    else if (arguments.count() > 1 && arguments.at(1).compare("update", Qt::CaseInsensitive) == 0) {
        table = RDFS_UPDATE;
    }

    Triple search;
    search.subject = search.object = QString();
    search.predicate = predicate;

    RdfsCursor *main = framework()->store()->cursor(search, table);
    if (!main->open()) {
        RdfsLogger::instance()->error(tr("Predicate is not used: %1").arg(predicate));
        delete main;
        return false; // Property not used
    }

    QSet<Triple> toRemove, processed;

    while (!main->eof()) {
        if (processed.contains(main->current())) {
            main->next();
            continue;
        }

        // Prepare stack
        QList<QString> atomList;
        QStack< QList<QString> > q;
        atomList.append(main->current().subject);
        atomList.append(main->current().object);
        q.push(atomList);
        processed << main->current();

        RdfsLogger::instance()->debug(tr("Processing: %1, %2, %3").arg(main->current().subject, main->current().predicate, main->current().object));

        // Find all possible cycles starting at current triple
        while (!q.isEmpty()) {
            atomList = q.pop();
            QString atom = atomList.last();

            // Find related triples at end of chain
            Triple search;
            search.subject = atom;
            search.predicate = predicate;
            search.object = QString();
            QList<Triple> results = framework()->store()->findTriples(search, table);

            foreach(Triple t, results) {
                // Check for cycle
                if (atomList.contains(t.object)) {
                    // Cycle! Remove triple, skip further search
                    toRemove << t;
                    processed << t;
                    continue;
                }
                if (processed.contains(t)) {
                    // Triple already being removed, or previously processed - skip
                    continue;
                }
                processed << t;

                // Expand list with t.object, push on stack
                QList<QString> newAtomList(atomList);
                newAtomList.append(t.object);
                q.push(newAtomList);
            }
        }
        RdfsLogger::instance()->debug(tr("Processed %1 triple(s) in total").arg(processed.size()));
        main->next();
    }

    main->close();
    delete main;

    // Remove triples
    RdfsLogger::instance()->log(tr("Removing %1 triples...").arg(toRemove.size()));

    Q_FOREACH(Triple t, toRemove) {
        RdfsLogger::instance()->debug(tr("Triple %1 %2 %3").arg(t.subject, t.predicate, t.object));
        framework()->store()->removeTriple(t, table);
    }
    return true;
}

QString RdfsKillCycles::checkArguments(const QStringList& arguments) const
{
    if (arguments.at(0).isEmpty()) {
        return tr("Empty predicate given.");
    }

    if (arguments.count() == 2 && arguments.at(1).compare("original", Qt::CaseInsensitive) != 0 &&
        arguments.at(1).compare("update", Qt::CaseInsensitive) != 0) {
            return tr("Unknown table %1 given.").arg(arguments.at(1));
    }

    return QString();
}

QList< Rdfs::CommandParameter > RdfsKillCycles::arguments() const
{
    Rdfs::CommandParameter predicate;
    predicate.name = tr("predicate");
    predicate.required = true;
    predicate.help = tr("The transitive predicate to check for cycles.");
    predicate.examples = QStringList() << "sc" << "sp";

    Rdfs::CommandParameter table;
    table.name = tr("table");
    table.required = false;
    table.help = tr("The graph name (original, update) to get the statistics from.");
    table.examples = QStringList() << "original" << "update";

    return QList<CommandParameter>() << predicate << table ;
}

RdfsKillCycles::RdfsKillCycles(Rdfs::RdfsUpdate* parent): RdfsCommand(parent)
{

}

RdfsKillCycles::~RdfsKillCycles()
{

}

#include "rdfskillcycles.moc"
