/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "rdfsontologydot.h"

#include <QFile>
#include <QFileInfo>
#include <QTextStream>
#include <stdio.h>

#include "../rdfsupdate.h"
#include "../rdfsbackend.h"
#include "../rdfscursor.h"
#include "../triple.h"
#include "../rdfslogger.h"

using namespace Rdfs;

bool RdfsOntologyDot::exec(const QStringList& arguments)
{
    QString propType = (arguments.size() > 1) ? arguments.at(1) : "Property";
    RdfsTable table = (arguments.size() > 2)
                    ? ((arguments.at(2).compare("original", Qt::CaseInsensitive) == 0)
                       ? RDFS_ORIGINAL
                       : RDFS_UPDATE )
                    : RDFS_ORIGINAL;

    QFile f;
    if (arguments.at(0) != "-") {
        f.setFileName(arguments.at(0));
        f.open(QIODevice::Truncate | QIODevice::WriteOnly | QIODevice::Text);
    }
    else {
        // Open stdout -> bypass RdfsLogger class preferrably
        f.open(stdout, QIODevice::WriteOnly | QIODevice::Text);
    }
    QTextStream t(&f);

    // List properties
    quint64 count = 0;
    t << "digraph " << propType << " {" << endl;

    Triple tt;
    tt.subject = QString();
    tt.predicate = "type";
    tt.object = propType;

    Rdfs::RdfsCursor *c = framework()->store()->cursor(tt);
    if (c->open()) {
        while (!c->eof()) {
            tt = c->current();
            // Find domain and range - assuming there's only one of each!
            Triple tdomain, trange;
            QList<Triple> results;
            tdomain.subject = tt.subject;
            tdomain.predicate = "dom";
            tdomain.object = QString();
            results = framework()->store()->findTriples(tdomain, table);
            if (results.isEmpty()) {
                c->next();
                continue; // Skip this property - no domain
            }
            tdomain = results.at(0);

            trange.subject = tt.subject;
            trange.predicate = "range";
            trange.object = QString();
            results = framework()->store()->findTriples(trange, table);
            if (results.isEmpty()) {
                c->next();
                continue; // Skip this property - no range
            }
            trange = results.at(0);

            if (tdomain.object.isEmpty() || trange.object.isEmpty()) {
                c->next();
                continue; // Skip this property - empty domain or range
            }

            t << tdomain.object << " -> " << trange.object << " [label=\"" << tt.subject << "\"]" << endl;
            c->next();
            ++count;
        }
        c->close();
    }
    t << "}";
    delete c;

    RdfsLogger::instance()->log(tr("Graph %1 with %2 edges written to file %3").arg(propType).arg(count).arg(arguments.at(0)));
    return true;
}

QString RdfsOntologyDot::checkArguments(const QStringList& arguments) const
{
    if (arguments.at(0) != "-") {
        // Check if file is writable
        QFileInfo file(arguments.at(0));
        QFileInfo dir(file.path());

        if (file.isDir()) {
            return tr("File %1 is a directory.").arg(arguments.at(0));
        }
        else if (file.exists() && !file.isWritable()) {
            return tr("File %1 is not writable.").arg(arguments.at(0));
        }
        else if (!file.exists() && (!dir.exists() || !dir.isDir() || !dir.isWritable())) {
            return tr("Cannot create file %1.").arg(arguments.at(0));
        }
    }

    if (arguments.size() == 3) {
        if (arguments.at(2).compare("original", Qt::CaseInsensitive) != 0 &&
           (arguments.at(2).compare("update", Qt::CaseInsensitive) != 0)) {
            return tr("Unknown table %1 given.").arg(arguments.at(2));
        }
    }
    return QString();
}

QList< Rdfs::CommandParameter > RdfsOntologyDot::arguments() const
{
    Rdfs::CommandParameter file, propType, table;
    file.name = tr("file");
    file.help = tr("The filename to write to.");
    file.examples << "graph.dot";
    file.required = true;

    propType.name = tr("proptype");
    propType.help = tr("The type designated to properties.");
    propType.examples << "Property";
    propType.required = false;

    table.name = tr("table");
    table.examples << "original";
    table.examples << "update";
    table.help = tr("Table to use for this command");
    table.required = false;

    return QList<CommandParameter>() << file << propType << table;
}

RdfsOntologyDot::RdfsOntologyDot(Rdfs::RdfsUpdate* parent): RdfsCommand(parent)
{

}

RdfsOntologyDot::~RdfsOntologyDot()
{

}

#include "rdfsontologydot.moc"
