/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "rdfssqlconnect.h"

#include <QUrl>
#include <QCoreApplication>
#include <QSettings>
#include <QDir>
#include <QFileInfo>
#ifdef WITH_SQL
    #include <QSqlError>
#endif

#include "../rdfsupdate.h"
#include "../rdfslogger.h"
#include "../rdfsbackend.h"

#ifdef WITH_SQL
    #include "../rdfssql.h"
    #include "../rdfsmysql.h"
    #include "../rdfssqlite.h"
#endif
#ifdef WITH_SOPRANO
    #include "../rdfsvirtuoso.h"
#endif
#ifdef WITH_BDB
    #include "../rdfsbdb.h"
    #include "../rdfsbdbbt.h"
    #include "../rdfsbdbtriplet.h"
#endif

using namespace Rdfs;

RdfsSqlConnect::RdfsSqlConnect(RdfsUpdate* parent)
    : RdfsCommand(parent)
{

}

QList<CommandParameter> RdfsSqlConnect::arguments() const
{
    CommandParameter url;
    url.name = tr("url");
#ifdef WITH_SQL
    url.examples << tr("mysql://username:password@host:port/database");
    url.examples << tr("mysql://username:password@localhost/database?socket=/var/run/mysql/mysqld.sock");
    url.examples << tr("sqlite:///path/to/file");
#endif
#ifdef WITH_SOPRANO
    url.examples << tr("virtuoso:///path/to/location");
    url.examples << tr("virtuoso://user:password@host:port");
#endif
#ifdef WITH_BDB
    url.examples << tr("bdb:///path/to/filename.db");
    url.examples << tr("bdb:///path/to/directory/");
    url.examples << tr("bdb:///path/to/primary.db?incoming=secondary1.db&outgoing=secondary2.db&edge=secondary3.db&cache=256k&pagesize=4k");

    url.examples << tr("bdbbt:///path/to/filename.db");
    url.examples << tr("bdbbt:///path/to/directory/");
    url.examples << tr("bdbbt:///path/to/original.db?update=update.db&rules=rules.db&cache=256k&pagesize=4k");

    url.examples << tr("bdbtt:///path/to/filename.db");
    url.examples << tr("bdbtt:///path/to/directory/");
    url.examples << tr("bdbtt:///path/to/original.db?update=update.db&rules=rules.db&cache=256k&pagesize=4k");
#endif
    url.help = tr("Url which points to a database location. See the examples for supported back-ends and options. Leave empty to reconnect using the last used connection.");
    url.required = false;

    return (QList<CommandParameter>() << url);
}

QString RdfsSqlConnect::checkArguments(const QStringList &arguments) const {
    QStringList schemes;
#ifdef WITH_SQL
    schemes << "mysql";
    schemes << "sqlite";
#endif
#ifdef WITH_SOPRANO
    schemes << "virtuoso";
#endif
#ifdef WITH_BDB
    schemes << "bdb";
    schemes << "bdbbt";
    schemes << "bdbtt";
#endif

    if (arguments.size() > 0) {
        QUrl uri;

        uri.setUrl(arguments.at(0));
        RdfsLogger::instance()->debug("Scheme: " + uri.scheme());

        if (!schemes.contains(uri.scheme())) {
            return tr("Please enter a valid connection URL. Supported schemes are: %1.").arg(schemes.join(", "));
        }
    }
    else {
        // Make sure setting exists
        QString uriString;
        QSettings s;
        uriString = s.value("connection/url", QString()).toString();
        if (uriString.isEmpty()) {
            return tr("No previous connection found. Use \"help\" for usage.");
        }
        QUrl uri;
        uri.setUrl(uriString);

        if (!schemes.contains(uri.scheme())) {
            return tr("Last saved connection is invalid. Supported schemes are: %1.").arg(schemes.join(", "));
        }
    }
    return QString();
}

bool RdfsSqlConnect::exec(const QStringList &arguments) {
    QUrl uri;
    QString uriString;
    RdfsBackend *backend;
    QSettings s;

    uriString = (arguments.size() > 0) ? arguments.at(0) : s.value("connection/url", QString()).toString();
    uri.setUrl(uriString);

    if (arguments.isEmpty()) {
        RdfsLogger::instance()->log(tr("Connecting to: %1").arg(uriString));
    }

    QString dbType, dbName, dbHost, dbSocket, dbPrimary, dbIncoming, dbOutgoing, dbEdge, dbOriginal, dbUpdate, dbRules;
    int dbPort;
#ifdef WITH_SQL
    if (uri.scheme().compare("mysql", Qt::CaseInsensitive) == 0) {
        // Connect to MySQL
        dbType = "MySQL";
        dbHost = uri.host();
        dbName = uri.path().split('/', QString::SkipEmptyParts).at(0);
        dbPort = uri.port(3306);
        if (uri.hasQueryItem("socket")) {
            dbSocket = uri.queryItemValue("socket");
        }

        backend = new RdfsMySql(dbName, uri.host(), dbPort, uri.userName(), uri.password(), dbSocket);
    }
    else if (uri.scheme().compare("sqlite", Qt::CaseInsensitive) == 0) {
        dbType = "SQLite3";
        dbName = uri.path();
        backend = new RdfsSqlite(uri.path());
    }
#endif

#ifdef WITH_SOPRANO
    if (uri.scheme().compare("virtuoso", Qt::CaseInsensitive) == 0) {
        dbType = "Virtuoso";
        dbHost = uri.host();
        dbName = uri.path();
        dbPort = uri.port(0);
        if (uri.hasQueryItem("socket")) {
            dbSocket = uri.queryItemValue("socket");
        }

        if (!dbHost.isEmpty()) {
            backend = new RdfsVirtuoso(dbHost, dbPort, uri.userName(), uri.password());
        }
        else {
            backend = new RdfsVirtuoso(dbName);
        }
    }
#endif

#ifdef WITH_BDB
    if (uri.scheme().compare("bdb", Qt::CaseInsensitive) == 0) {
        QString cacheSize, pageSize;

        dbType = "Berkeley";
        dbName = uri.path();
        QFileInfo dbInfo(dbName);
        if (dbInfo.isDir()) {
            dbName += QDir::separator() + QString("rdfsupdate.db");
            dbInfo.setFile(dbName);
        }
        if (dbInfo.fileName().isEmpty()) {
            // Path ends in slash
            dbName += "rdfsupdate.db";
            dbInfo.setFile(dbName);
        }

        dbPrimary = dbInfo.fileName();
        dbIncoming = (uri.hasQueryItem("incoming"))
                   ? uri.queryItemValue("incoming")
                   : dbInfo.baseName() + "-incoming." + dbInfo.completeSuffix();
        dbOutgoing = (uri.hasQueryItem("outgoing"))
                   ? uri.queryItemValue("outgoing")
                   : dbInfo.baseName() + "-outgoing." + dbInfo.completeSuffix();
        dbEdge     = (uri.hasQueryItem("edge"))
                   ? uri.queryItemValue("edge")
                   : dbInfo.baseName() + "-edge." + dbInfo.completeSuffix();
        cacheSize  = (uri.hasQueryItem("cache"))
                   ? uri.queryItemValue("cache")
                   : QString();
        pageSize   = (uri.hasQueryItem("pagesize"))
                   ? uri.queryItemValue("pagesize")
                   : QString();

        RdfsLogger::instance()->debug(tr("Secondary database: incoming: %1").arg(dbIncoming));
        RdfsLogger::instance()->debug(tr("Secondary database: outgoing: %1").arg(dbOutgoing));
        RdfsLogger::instance()->debug(tr("Secondary database: edge    : %1").arg(dbEdge));

        backend = new RdfsBdb(dbInfo.absolutePath(), dbPrimary, dbIncoming, dbOutgoing, dbEdge, cacheSize, pageSize);
    }

    // BTree
    if (uri.scheme().compare("bdbbt", Qt::CaseInsensitive) == 0) {
        QString cacheSize, pageSize;

        dbType = "BerkeleyBTree";
        dbName = uri.path();
        QFileInfo dbInfo(dbName);
        if (dbInfo.isDir()) {
            dbName += QDir::separator() + QString("original.db");
            dbInfo.setFile(dbName);
        }
        if (dbInfo.fileName().isEmpty()) {
            // Path ends in slash
            dbName += "original.db";
            dbInfo.setFile(dbName);
        }

        dbOriginal = dbInfo.fileName();
        dbUpdate = (uri.hasQueryItem("update"))
                   ? uri.queryItemValue("update")
                   : "update." + dbInfo.completeSuffix();
        dbRules = (uri.hasQueryItem("rules"))
                   ? uri.queryItemValue("rules")
                   : "rules." + dbInfo.completeSuffix();
        cacheSize  = (uri.hasQueryItem("cache"))
                   ? uri.queryItemValue("cache")
                   : QString();
        pageSize   = (uri.hasQueryItem("pagesize"))
                   ? uri.queryItemValue("pagesize")
                   : QString();

        RdfsLogger::instance()->debug(tr("Update database: %1").arg(dbUpdate));
        RdfsLogger::instance()->debug(tr("Rules database: %1").arg(dbRules));

        backend = new RdfsBdbBt(dbInfo.absolutePath(), dbOriginal, dbUpdate, dbRules, cacheSize, pageSize);
    }

    // TripleT
    if (uri.scheme().compare("bdbtt", Qt::CaseInsensitive) == 0) {
        QString cacheSize, pageSize;

        dbType = "BerkeleyTripleT";
        dbName = uri.path();
        QFileInfo dbInfo(dbName);
        if (dbInfo.isDir()) {
            dbName += QDir::separator() + QString("original.db");
            dbInfo.setFile(dbName);
        }
        if (dbInfo.fileName().isEmpty()) {
            // Path ends in slash
            dbName += "original.db";
            dbInfo.setFile(dbName);
        }

        dbOriginal = dbInfo.fileName();
        dbUpdate = (uri.hasQueryItem("update"))
                   ? uri.queryItemValue("update")
                   : "update." + dbInfo.completeSuffix();
        dbRules = (uri.hasQueryItem("rules"))
                   ? uri.queryItemValue("rules")
                   : "rules." + dbInfo.completeSuffix();
        cacheSize  = (uri.hasQueryItem("cache"))
                   ? uri.queryItemValue("cache")
                   : QString();
        pageSize   = (uri.hasQueryItem("pagesize"))
                   ? uri.queryItemValue("pagesize")
                   : QString();

        RdfsLogger::instance()->debug(tr("Update database: %1").arg(dbUpdate));
        RdfsLogger::instance()->debug(tr("Rules database: %1").arg(dbRules));

        backend = new RdfsBdbTripleT(dbInfo.absolutePath(), dbOriginal, dbUpdate, dbRules, cacheSize, pageSize);
    }
#endif

    if (backend->isOpen()) {
        if (dbHost.isEmpty()) {
            RdfsLogger::instance()->log(tr("Successfully connected to %1 database %2").
                                           arg(dbType, dbName));
        }
        else {
            RdfsLogger::instance()->log(tr("Successfully connected to %1 database %2@%3").
                                           arg(dbType, dbName, dbHost));
        }
        // Delete previous connection
        framework()->disconnectStore();
        // Set new connection
        framework()->setStore(backend);
        // Store settings
        s.setValue("connection/url", uriString);
        return true;
    }
    else {
#ifdef WITH_SQL
        // TODO What if not connected to SQL despite WITH_SQL?
        RdfsSql *database = dynamic_cast<RdfsSql*>(backend);
        if (database) {
            RdfsLogger::instance()->error(tr("Error connecting to %1 database %2:\n"
                                            "Driver error: %3\n"
                                            "Database error: %4")
                                        .arg(dbType, dbName, database->lastError().driverText(), database->lastError().databaseText()));
        }
        else {
#endif
            // Error for non-sql back-ends
            RdfsLogger::instance()->error(tr("Error creating connection to backend"));
#ifdef WITH_SQL
        }
#endif

        // Keep existing connection in framework
        delete backend;
        return false;
    }
}

RdfsSqlConnect::~RdfsSqlConnect() {

}

#include "rdfssqlconnect.moc"
