/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "rdfsunchase.h"

#include <QQueue>
#include <QSet>
#include <QtDebug>

#include "rdfsclosurecheck.h"
#include "../rdfsupdate.h"
#include "../rdfssql.h"
#include "../rdfslogger.h"

using namespace Rdfs;

RdfsUnchase::RdfsUnchase(RdfsUpdate* parent)
    : RdfsTripleCommand(parent) {
}

bool RdfsUnchase::exec(const QStringList& arguments) {
    Triple t;
    RdfsTable table;
    extractArguments(arguments, &table, &t);

    m_boundVarCount = 0;

    QSet<Triple> triples = unchase(t, framework()->store()->rules(), table).keys().toSet();

    // Log results, if any
    if (!triples.isEmpty()) {
        RdfsLogger::instance()->log(tr("Unchase for %1 %2 %3:").arg(t.subject, t.predicate, t.object));
        Q_FOREACH(Triple tt, triples) {
            RdfsLogger::instance()->log(QString("%1 %2 %3").arg(tt.subject, tt.predicate, tt.object));
        }
    }
    else {
        RdfsLogger::instance()->log(tr("No unchase for: %1 %2 %3").arg(t.subject, t.predicate, t.object));
    }
    return true;
}

QMap<Triple, VariableInstances> RdfsUnchase::unchase(Triple t, QList<Rule> rules, RdfsTable table, QSet<Triple> processed) {
    const QRegExp varRegex("@[0-9]+");

    RdfsLogger::instance()->debug(QString("Unchasing: (%1, %2, %3)").arg(t.subject, t.predicate, t.object));

    QMap<Triple, VariableInstances> result;

    // Add current triple to processed list
    processed << clearVars(t);

    // Get the source candidates, remove previously processed items (including current)
    QMap<Triple, VariableInstances> s = sources(t, rules, table);

    // Remove the set of already processed triples
    Q_FOREACH(Triple tt, processed) {
        s.remove(tt);
    }

    // Of all sources: get their deterministic unchase and merge the result
    // Note that we also want to unchase any instances retrieved earlier,
    QMapIterator<Triple, VariableInstances> i(s);
    while (i.hasNext()) {
        i.next();

        Triple tt = i.key();
        VariableInstances vars = i.value();

        // Rename variables in tt: create new global bindings for each local variable
        // still left in tt
        if (varRegex.exactMatch(tt.subject)) {
            tt.subject = "$" + QString::number(m_boundVarCount);
            m_boundVarCount++;
        }
        if (varRegex.exactMatch(tt.predicate)) {
            tt.predicate = "$" + QString::number(m_boundVarCount);
            m_boundVarCount++;
        }
        if (varRegex.exactMatch(tt.object)) {
            tt.object = "$" + QString::number(m_boundVarCount);
            m_boundVarCount++;
        }

        // Ensure that triple is not already processed
        if (processed.contains(clearVars(tt))) {
            continue;
        }

        QMap<Triple, VariableInstances> sourceUnchase = unchase(tt, rules, table, processed);

        // Loop over the results, and merge them. Also merge the variable instances
        QMapIterator<Triple, VariableInstances> j(sourceUnchase);
        while (j.hasNext()) {
            j.next();
            if (!result.contains(j.key())) {
                // Note that bindings will be on unique variables, so no duplicates

                VariableInstances currentVars = mergeVarInstances(vars, j.value());
                result.insert(j.key(), currentVars);

                // Instantiate current triple and add it with the variable bindings as well
                result.insert(instantiate(t, currentVars), currentVars);
                RdfsLogger::instance()->debug(QString("Inst: (%1, %2, %3), with: %4").arg(tt.subject, tt.predicate, tt.object, variableMapToString(currentVars)));
            }
        }
    }

    // Check the graph for matches of the triple, add these to the result
    // Note that there's no need to unchase these. Since they are instances of t,
    // if the unchase is larger than just this instance, that means rules were applied.
    // The original triple t will then also match these rules.
    QMap<Triple, VariableInstances> gi = tripleInstances(t, table);
    // Unite these instances with the result. Note tnat QMap::unite() causes duplicates, so
    // we avoid that
    QMapIterator<Triple, VariableInstances> j(gi);
    while (j.hasNext()) {
        j.next();
        if (!result.contains(j.key())) {
            result.insert(j.key(), j.value());
        }
    }

    // And the result is done :)
    return result;
}

QMap<Triple, VariableInstances> RdfsUnchase::sources(Triple t, QList<Rule> rules, RdfsTable table) {
    Triple tt;
    QMap<Triple, VariableInstances> sources;
    VariableInstances varBindings;
    VariableInstances boundVarInsts;

    Q_FOREACH(Rule r, rules) {
        varBindings.clear();
        boundVarInsts.clear();

        // Bind second triple in rule to given triple t, continue with next rule
        // if there is no match
        bool match = (checkBinding(t.subject, r.second.subject, &varBindings, &boundVarInsts) &&
            checkBinding(t.predicate, r.second.predicate, &varBindings, &boundVarInsts) &&
            checkBinding(t.object, r.second.object, &varBindings, &boundVarInsts));

        // If no valid match: continue with next rule
        if (!match) {
            continue;
        }

        // Fill in source triple of rule
        Triple tt;
        tt.subject = (varBindings.contains(r.first.subject)) ? varBindings.value(r.first.subject) : r.first.subject;
        tt.predicate = (varBindings.contains(r.first.predicate)) ? varBindings.value(r.first.predicate) : r.first.predicate;
        tt.object = (varBindings.contains(r.first.object)) ? varBindings.value(r.first.object) : r.first.object;

        // Now instantiate as much as possible from tt
        tt = instantiate(tt, boundVarInsts);

        // Add tt to result list
        sources.insert(tt, boundVarInsts);

        RdfsLogger::instance()->debug(QString("Got source for (%1, %2, %3): (%4, %5, %6).").
                arg(t.subject, t.predicate, t.object, tt.subject, tt.predicate, tt.object));
    }
    return sources;
}

QMap<Triple, VariableInstances> RdfsUnchase::tripleInstances(Triple t, RdfsTable table) {
    // Prepare triple t with blanks
    QMap<Triple, VariableInstances> result;
    Triple tWithBlanks;
    const QRegExp boundVarRegex("\\$[0-9]+");

    if (!boundVarRegex.exactMatch(t.subject)) {
        tWithBlanks.subject = t.subject;
    }
    if (!boundVarRegex.exactMatch(t.predicate)) {
        tWithBlanks.predicate = t.predicate;
    }
    if (!boundVarRegex.exactMatch(t.object)) {
        tWithBlanks.object = t.object;
    }

    QSet<Triple> triples = framework()->store()->findTriples(tWithBlanks, table).toSet();

    // Note that we must check that the instantiations are valid. If the same
    // variable is in t twice, the values must match
    VariableInstances boundVars;
    Q_FOREACH(Triple tt, triples) {
        boundVars.clear();
        bool match = (checkBinding(t.subject, tt.subject, 0, &boundVars) &&
            checkBinding(t.predicate, tt.predicate, 0, &boundVars) &&
            checkBinding(t.object, tt.object, 0, &boundVars));

        if (match) {
            result.insert(tt, boundVars);
        }
    }
    return result;
}

bool RdfsUnchase::checkBinding(QString t1, QString t2, VariableInstances* localVars, VariableInstances* boundVars)
{
    const QRegExp varRegex("@[0-9]+");
    const QRegExp boundVarRegex("\\$[0-9]+");

    // If the triple to be matched has a variable here...
    // Note: "local vars" may be 0-pointer, if we're not trying to bind a triple from a rule
    if (localVars != 0) {
        if (varRegex.exactMatch(t2)) {
            if (!localVars->contains(t2)) {
                // New local (rule) variable: bind value
                localVars->insert(t2, t1);
                return true;
            }
            if (t1 == localVars->value(t2)) {
                // Local variable already bound, given value matches
                return true;
            }

            // Check if bound value of local variable is a global variable
            if (boundVarRegex.exactMatch(localVars->value(t2))) {
                // Check global variables for possible instantiation
                if (boundVars->contains(localVars->value(t2))) {
                    // Make sure current global instance matches t1. If not, the instantiation is not valid
                    return (boundVars->value(localVars->value(t2)) == t1);
                }
                else {
                    // Create global binding
                    RdfsLogger::instance()->debug(QString("Binding %1 to %2").arg(localVars->value(t2), t1));
                    boundVars->insert(localVars->value(t2), t1);
                    return true;
                }
            }

            // Variable already bound, but no possible match
            return false;

        }
    }
    // From here on: t2 is a non-variable

    // If the matching triple has a variable here
    if (boundVarRegex.exactMatch(t1)) {
        if (boundVars->contains(t1)) {
            // Check that binding is the same
            return (boundVars->value(t1) == t2);
        }
        else {
            // Bind variable
            RdfsLogger::instance()->debug(QString("Binding %1 to %2").arg(t1, t2));
            boundVars->insert(t1, t2);
            return true;
        }
    }

    // Both non-variables: check equality
    return (t1 == t2);
}

Triple RdfsUnchase::instantiate(Triple t, VariableInstances vars)
{
    Triple tt(t);
    const QRegExp boundVarRegex("\\$[0-9]+");

    while (vars.contains(tt.subject)) {
        tt.subject = vars.value(tt.subject);
    }

    while (vars.contains(tt.predicate)) {
        tt.predicate = vars.value(tt.predicate);
    }

    while (vars.contains(tt.object)) {
        tt.object = vars.value(tt.object);
    }
    return tt;
}

Triple RdfsUnchase::clearVars(Triple t)
{
    const QRegExp boundVarRegex("\\$[0-9]+");
    Triple tt;
    if (!boundVarRegex.exactMatch(t.subject)) {
        tt.subject = t.subject;
    }

    if (!boundVarRegex.exactMatch(t.predicate)) {
        tt.predicate = t.predicate;
    }

    if (!boundVarRegex.exactMatch(t.object)) {
        tt.object = t.object;
    }

    return tt;
}


VariableInstances RdfsUnchase::mergeVarInstances(VariableInstances i1, VariableInstances i2)
{
    RdfsLogger::instance()->debug(QString("Merging {%1} with {%2}").arg(variableMapToString(i1), variableMapToString(i2)));
    VariableInstances result(i1);
    QMapIterator<QString, QString> i(i2);
    while (i.hasNext()) {
        i.next();
        if (!result.contains(i.key())) {
            result.insert(i.key(), i.value());
        }
    }
    return result;
}


#ifndef NDEBUG
QString RdfsUnchase::variableMapToString(VariableInstances instances)
{
    QStringList result;
    QMapIterator<QString, QString> i(instances);
    while (i.hasNext()) {
        i.next();
        result << "\"" + i.key() + "\" -> \"" + i.value() + "\"";
    }
    return result.join(", ");
}
#else
QString RdfsUnchase::variableMapToString(VariableInstances instances)
{
    return QString();
}
#endif

QString RdfsUnchase::checkArguments(const QStringList& arguments) const {
    QString baseResult = RdfsTripleCommand::checkArguments(arguments);
    if (baseResult.isEmpty()) {
        // Check that triple does not contain empty items
        RdfsTable table;
        Triple t;
        extractArguments(arguments, &table, &t);

        if (t.subject.isEmpty() || t.predicate.isEmpty() || t.object.isEmpty()) {
            return tr("The triple contains an empty value. This is not supported.");
        }
    }
    return baseResult;
}

QStringList RdfsUnchase::aliases() const {
    return QStringList() << "uc";
}

QString RdfsUnchase::description() const {
    return tr("Get the set of triples in the unchase of a given triple. Each returned triple is checked to be in the entailment of the graph.");
}

QString RdfsUnchase::name() const {
    return "unchase";
}

RdfsUnchase::~RdfsUnchase() {
}

#include "rdfsunchase.moc"
