/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "rdfsbdbbtcursor.h"

#include <stdlib.h>
#include "rdfslogger.h"

using namespace Rdfs;

RdfsBdbBtCursor::RdfsBdbBtCursor(RdfsBdbBt* backend, const Rdfs::Triple& query, RdfsTable table, DB** indexes)
    : RdfsCursor(backend, query, table)
{
    memset(&m_current, 0, sizeof(DBT));
    m_cursor = NULL;
    for (int i = 0; i < 3; ++i) {
        m_indexes[i] = indexes[i];
    }
    m_eof = true;
    m_open = false;
}

RdfsBdbBtCursor::~RdfsBdbBtCursor()
{
    close();
}


bool RdfsBdbBtCursor::eof() const
{
    return m_eof;
}

void RdfsBdbBtCursor::close()
{
    if (m_cursor != NULL) {
        m_cursor->c_close(m_cursor);
        m_cursor = NULL;
    }

    m_eof = true;
    m_open = false;
}

bool RdfsBdbBtCursor::open()
{
    m_cursor = findCursor(queryTriple(), m_indexes, &m_order);
    if (m_cursor == NULL) {
        m_eof = true;
        m_open = false;
        return false;
    }

    // Load current key
    DBT data;
    memset(&data, 0, sizeof(DBT));
    m_cursor->c_get(m_cursor, &m_current, &data, DB_CURRENT);

    m_currentTriple = RdfsBdbBt::dbtToTriple(&m_current, m_order);

    m_eof = false;
    moveValid();

    // Check if we have a result
    m_open = !eof();
    if (!m_open) {
        close();
    }

    return m_open;
}

void RdfsBdbBtCursor::advanceCursor()
{
    DBT data;
    memset(&data, 0, sizeof(DBT));

    int ret = m_cursor->c_get(m_cursor, &m_current, &data, DB_NEXT);
    if (ret != 0) {
        m_eof = true;
        // Clear triple
        m_currentTriple.subject = QString();
        m_currentTriple.predicate = QString();
        m_currentTriple.object = QString();
        return;
    }
    m_currentTriple = RdfsBdbBt::dbtToTriple(&m_current, m_order);
}


bool RdfsBdbBtCursor::next()
{
    advanceCursor();
    return moveValid();
}

Rdfs::Triple RdfsBdbBtCursor::current() const
{
    return m_currentTriple;
}

bool RdfsBdbBtCursor::isMatch() const
{
    Triple q = queryTriple();
    return ((q.subject.isEmpty() || q.subject == m_currentTriple.subject) &&
            (q.predicate.isEmpty() || q.predicate == m_currentTriple.predicate) &&
            (q.object.isEmpty() || q.object == m_currentTriple.object));
}

bool RdfsBdbBtCursor::moveValid()
{
    if (eof()) {
        return false;
    }

    while (!isMatch()) {
        // Check if advancing cursor may yield new triple
        if (!mayHaveMore()) {
            m_eof = true; // Found all matches
            return false;
        }

        // Move on
        advanceCursor();

        if (eof()) {
            return false;
        }
    }
    // Found a match
    return true;
}

bool RdfsBdbBtCursor::mayHaveMore()
{
    QString queryStrings[] = { queryTriple().subject, queryTriple().predicate, queryTriple().object };
    QString currentStrings[] = { m_currentTriple.subject, m_currentTriple.predicate, m_currentTriple.object };
    int indexes[3];

    switch(m_order) {
        case RdfsBdbBt::SOP:
            indexes[0] = 0;
            indexes[1] = 2;
            indexes[2] = 1;
            break;
        case RdfsBdbBt::PSO:
            indexes[0] = 1;
            indexes[1] = 0;
            indexes[2] = 2;
            break;
        case RdfsBdbBt::OSP:
            indexes[0] = 2;
            indexes[1] = 0;
            indexes[2] = 1;
            break;
    }

    // Check index order
    for (int i = 0; i < 3; ++i) {
        int index = indexes[i];
        // If query item is empty: we may have to skip some records to find a match, but there may be one
        // in the future
        if (queryStrings[index].isEmpty()) {
            return true;
        }
        // If query is not empty, and not a match anymore, we can stop
        else if (queryStrings[index] != currentStrings[index]) {
            return false;
        }
        // Else query is not empty, but matches: look into next index item
    }
    // All three index items match: continue search
    return true;
}


DBC* RdfsBdbBtCursor::findCursor(const Rdfs::Triple& t, DB** database, RdfsBdbBt::INDEX_ORDER* order) const
{
    // SOP / PSO / OSP
    if ((!t.subject.isEmpty() && !t.object.isEmpty()) || // Predicate unknown, or all three known
        (!t.subject.isEmpty() && t.predicate.isEmpty() && t.object.isEmpty()) || // Subject only known
        (t.subject.isEmpty() && t.predicate.isEmpty() && t.object.isEmpty())) { // All three unknown
        *order = RdfsBdbBt::SOP;
    }
    else if ((!t.subject.isEmpty() && !t.predicate.isEmpty()) || // Object unknown; if !object.isEmpty(), then SOP case would already match
        (t.subject.isEmpty() && !t.predicate.isEmpty() && t.object.isEmpty())) { // Predicate only known
        *order = RdfsBdbBt::PSO;
    }
    else { // Object only known, or subject only unknown
        *order = RdfsBdbBt::OSP;
    }

    return loadCursor(t, database, *order);
}

DBC* RdfsBdbBtCursor::loadCursor(const Rdfs::Triple& t, DB** database, const Rdfs::RdfsBdbBt::INDEX_ORDER& order) const
{
    DB *db = database[order];
    DBT key = RdfsBdbBt::tripleToDbt(t, order);
    DBT data;
    memset(&data, 0, sizeof(DBT));

    DBC* cursor;

    int ret;
    u_int32_t flags, cflags;
    flags = cflags = 0;
#ifdef DB_BULK
    // Add BULK cursor to prevent unnecessary cache misses
    flags = DB_BULK;
#endif

    // Init cursor
    if ((ret = db->cursor(db, NULL, &cursor, flags)) != 0) {
        free(key.data);
        return NULL;
    }
    void *dataptr = key.data;

    // If one atom is empty: do range traversal
    cflags = (t.subject.isEmpty() || t.predicate.isEmpty() || t.object.isEmpty())
           ? DB_SET_RANGE
           : DB_SET;
    // Set cursor to correct position
    ret = cursor->c_get(cursor, &key, &data, cflags);
    // Free allocated memory - the key has been overwritten
    free(dataptr);

    if (ret == 0) {
        return cursor;
    }
    else {
        cursor->c_close(cursor);
        return NULL;
    }
}

void RdfsBdbBtCursor::removeCurrent()
{
    if (eof()) {
        return;
    }

    // Find item in all orders
    DBC *c1, *c2;
    // Load other 2 cursors
    switch(m_order) {
        default:
        case RdfsBdbBt::SOP:
            c1 = loadCursor(m_currentTriple, m_indexes, RdfsBdbBt::PSO);
            c2 = loadCursor(m_currentTriple, m_indexes, RdfsBdbBt::OSP);
            break;
        case RdfsBdbBt::PSO:
            c1 = loadCursor(m_currentTriple, m_indexes, RdfsBdbBt::SOP);
            c2 = loadCursor(m_currentTriple, m_indexes, RdfsBdbBt::OSP);
            break;
        case RdfsBdbBt::OSP:
            c1 = loadCursor(m_currentTriple, m_indexes, RdfsBdbBt::SOP);
            c2 = loadCursor(m_currentTriple, m_indexes, RdfsBdbBt::PSO);
            break;
    }

    // If set, remove item from cursors
    if (c1 != NULL) {
        c1->c_del(c1, 0);
        c1->c_close(c1);
    }
    if (c2 != NULL) {
        c2->c_del(c2, 0);
        c2->c_close(c2);
    }

    // Delete from own index, and move cursor to next item
    m_cursor->c_del(m_cursor, 0);
    next();
}
