/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef RDFS_RDFSBDBBTCURSOR_H
#define RDFS_RDFSBDBBTCURSOR_H

#include "rdfscursor.h"

#include "rdfsbdbbt.h"

namespace Rdfs {

class RdfsBdbBtCursor : public Rdfs::RdfsCursor
{
private:
    DBC *m_cursor;
    DB* m_indexes[3];
    DBT m_current;
    Triple m_currentTriple;
    RdfsBdbBt::INDEX_ORDER m_order;
    bool m_eof;

    /**
     * @brief Determine the best index to use, and open a cursor for the given query.
     *
     * @param t The triple to find.
     * @param database The database set to use.
     * @param order The selected order will be stored in this variable
     * @return DBC* The cursor.
     **/
    DBC* findCursor(const Rdfs::Triple& t, DB** database, RdfsBdbBt::INDEX_ORDER *order) const;

    /**
     * @brief Load a cursor for a triple in a given order.
     *
     * @param t The triple to search
     * @param database The database set to search in
     * @param order The index order to use
     * @return DBC* The cursor
     **/
    DBC* loadCursor(const Rdfs::Triple& t, DB** database, const RdfsBdbBt::INDEX_ORDER &order) const;

    /**
     * @brief Check if current item is a match of the original query.
     *
     * Sometimes in the, depending on the query and index order, some items may occur which need to
     * be skipped.
     *
     * @return bool If the current item is a match of the original query or not.
     **/
    bool isMatch() const;


    /**
     * @brief Move the cursor to a position with a valid match of the query.
     *
     * If the current position of the cursor matches
     * @return bool True if a match was found at the current or a next position, false if there are no
     *              more matches to be found.
     **/
    bool moveValid();

    /**
     * @brief Move the cursor to the next item.
     **/
    void advanceCursor();

    /**
     * Check if the cursor may come across more matches of the query, judging by the current
     * position.
     * @return Whether there may be more matches or not
     */
    bool mayHaveMore();

public:
    RdfsBdbBtCursor(Rdfs::RdfsBdbBt* backend, const Rdfs::Triple& query, Rdfs::RdfsTable table, DB** indexes);
    virtual ~RdfsBdbBtCursor();

    virtual Rdfs::Triple current() const;
    virtual bool open();
    virtual bool next();
    virtual bool eof() const;
    virtual void close();

    /**
     * @brief Remove the triple the cursor currently points to.
     *
     * This will remove the triple from the database entirely. This includes all three
     * indexes, not just the most optimal one.
     **/
    virtual void removeCurrent();
};

}

#endif // RDFS_RDFSBDBBTCURSOR_H
