/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef RDFS_RDFSBDBTRIPLET_H
#define RDFS_RDFSBDBTRIPLET_H

#include "rdfstripletbackend.h"

#include <db.h>

namespace Rdfs {

class BdbTripleTBtreeCursor;

class RdfsBdbTripleT : public Rdfs::RdfsTripleTBackend
{
private:
    friend class BdbTripleTBtreeCursor;

    DB_ENV *m_db_env; /**< Database environment */
    DB *m_db_original; /**< DB pointer for original table */
    DB *m_db_update;  /**< DB pointer for updated table */
    DB *m_db_rules;   /**< DB pointer for rules table */

    QString m_dir, m_file, m_updateFile, m_rulesFile;
    QMap<QString, DB*> m_btreesOriginal;
    QMap<QString, DB*> m_btreesUpdate;

    DBT payloadToDbt(QByteArray data);
    QByteArray dbtToPayload(DBT *data);

    DB *database(Rdfs::RdfsTable table);

    /**
     * Retrieve a specific BTree database/index. The atom and role describe the specific database to load, within the
     * original @p table.
     *
     * If @p updateBtreeIndex is true (default), the file with information about existing B-tree database
     * will be updated if a new index is created. As such, the index will be automatically loaded upon
     * creation of the main index.
     *
     * @param atom The atom of the index to load.
     * @param role The role of the atom
     * @param table The original table of the storage
     * @param updateBtreeIndex Whether to update the B-tree list file.
     *
     * @return The DB handle of the B-tree index. NULL if something failed.
     */
    DB *btreeDatabase(const QString &atom, const RdfsRole &role, const RdfsTable &table, bool updateBtreeIndex = TRUE);

    /**
     * Overload for btreeDatabase(), which is based on the pure index (atom + role) of the
     * index.
     */
    DB *btreeDatabase(const QString &id, const RdfsTable &table, bool updateBtreeIndex = TRUE);

    static DBT btreeDbt(const QString& secondary, const QString& tertiary);
    static Triple btreeTripleFromDbt(DBT *record, DBT *data, const QString& primary);

    int sizeStringToBytes(const QString &sizeString);

protected:
    virtual QStringList atoms(Rdfs::RdfsTable table = RDFS_ORIGINAL);
    virtual void putPayload(const QString &atom, const RdfsRole &role, QByteArray data, Rdfs::RdfsTable table = RDFS_ORIGINAL);
    virtual QByteArray payload(const QString &atom, const RdfsRole &role,  Rdfs::RdfsTable table = RDFS_ORIGINAL);

    virtual RdfsCursor* btreeCursor(const QString& atom, const Rdfs::RdfsRole& role, const QString& secondary, const QString& tertiary, RdfsTable table);
    virtual void insertBtreeTriple(const QString& atom, const Rdfs::RdfsRole& role, const QString& secondary, const QString& tertiary, unsigned char distance, RdfsTable table);
    virtual void removeBtreeTriple(const QString& atom, const Rdfs::RdfsRole& role, const QString& secondary, const QString& tertiary, RdfsTable table);

public:
    RdfsBdbTripleT(const QString& dir, const QString& file, const QString& updateFile, const QString& rulesFile,
                   const QString& cacheSize = QString(), const QString& pageSize = QString());
    virtual ~RdfsBdbTripleT();

    virtual void setPageSize(const QString &size);
    virtual void setPageSize(int bytes);
    virtual void setCacheSize(const QString &size);
    virtual void setCacheSize(int bytes);

    virtual void truncate(Rdfs::RdfsTable table = RDFS_ORIGINAL);
    virtual int capabilities() const;
    virtual void close();
    virtual bool isOpen();
    virtual Rdfs::CacheStats statistics() const;
};

}

#endif // RDFS_RDFSBDBTRIPLET_H
