/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "rdfsgraph.h"
#include "rdfslogger.h"

using namespace Rdfs;


RdfsGraph::RdfsGraph() {
    // Nothing to initialize
}

void RdfsGraph::clear() {
    m_edges.clear();
    m_nodes.clear();
    m_incomingEdges.clear();
    m_outgoingEdges.clear();
    m_nodeAttributes.clear();
    m_edgeAttributes.clear();
}

void RdfsGraph::add(Triple t) {
    m_edges << t;
    m_nodes << t.subject;
    m_nodes << t.object;

    // Administer incoming/outgoing edges
    QMap<QString, TripleSet>::iterator i;
    i = m_outgoingEdges.find(t.subject);
    if (i != m_outgoingEdges.end()) {
        i.value().insert(t);
    }
    else {
        TripleSet l;
        l << t;
        m_outgoingEdges.insert(t.subject, l);
    }

    i = m_incomingEdges.find(t.object);
    if (i != m_incomingEdges.end()) {
        i.value().insert(t);
    }
    else {
        TripleSet l;
        l << t;
        m_incomingEdges.insert(t.object, l);
    }
}

void RdfsGraph::add(QList<Triple> triples)
{
    Triple t;
    Q_FOREACH(t, triples) {
        add(t);
    }
}

void RdfsGraph::add(QSet<Triple> triples) {
    Triple t;
    Q_FOREACH(t, triples) {
        add(t);
    }
}

Rdfs::RdfsGraph operator<<(Rdfs::RdfsGraph &g, const Triple &t) {
    g.add(t);
    return (g);
}

void RdfsGraph::remove(Triple t) {
    // Remove edge from set
    if (!m_edges.contains(t)) {
        // Triple doesn't exist
        return;
    }
    m_edges.remove(t);

    // Administer incoming/outgoing edges
    QMap<QString, TripleSet>::iterator i;
    i = m_outgoingEdges.find(t.subject);
    i.value().remove(t);

    i = m_incomingEdges.find(t.object);
    i.value().remove(t);

    // Remove subject/object if no incoming/outgoing edges left (nodes disconnected)
    if (m_incomingEdges.value(t.subject).isEmpty() && m_outgoingEdges.value(t.subject).isEmpty()) {
        m_incomingEdges.remove(t.subject);
        m_outgoingEdges.remove(t.subject);
        // Remove attributes
        QMap<QString, NodeAttributeValues>::iterator a;
        for (a = m_nodeAttributes.begin(); a != m_nodeAttributes.end(); ++a) {
            a.value().remove(t.subject);
        }
    }
    if (m_incomingEdges.value(t.object).isEmpty() && m_outgoingEdges.value(t.object).isEmpty()) {
        m_incomingEdges.remove(t.object);
        m_outgoingEdges.remove(t.object);
        // Remove attributes
        QMap<QString, NodeAttributeValues>::iterator a;
        for (a = m_nodeAttributes.begin(); a != m_nodeAttributes.end(); ++a) {
            a.value().remove(t.object);
        }
    }

    // Remove attributes assigned to triple
    QMap<QString, EdgeAttributeValues>::iterator ea;
    for (ea = m_edgeAttributes.begin(); ea != m_edgeAttributes.end(); ++ea) {
        ea.value().remove(t);
    }
}

void RdfsGraph::setNodeAttribute(QString attribute, QString node, QString value) {
    if (!m_nodes.contains(node)) {
        // Can't add attribute to non-existing node
        return;
    }

    QMap<QString, NodeAttributeValues>::iterator i;
    i = m_nodeAttributes.find(attribute);
    if (i != m_nodeAttributes.end()) {
        i.value().insert(node, value);
    }
    else {
        NodeAttributeValues l;
        l.insert(node, value);
        m_nodeAttributes.insert(attribute, l);
    }
}

void RdfsGraph::setEdgeAttribute(QString attribute, Triple edge, QString value) {
    if (!m_edges.contains(edge)) {
        // Can't add attribute to non-existing edge
        return;
    }

    QMap<QString, EdgeAttributeValues>::iterator i;
    i = m_edgeAttributes.find(attribute);
    if (i != m_edgeAttributes.end()) {
        i.value().insert(edge, value);
    }
    else {
        EdgeAttributeValues l;
        l.insert(edge, value);
        m_edgeAttributes.insert(attribute, l);
    }
}

RdfsGraph::~RdfsGraph() {
}

