/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef RDFS_RDFSGRAPH_H
#define RDFS_RDFSGRAPH_H

#include "triple.h"

#include <QString>
#include <QList>
#include <QSet>
#include <QMap>
#include <QMultiMap>

namespace Rdfs {

typedef QMultiMap<Triple, QString>  EdgeAttributeValues;
typedef QMultiMap<QString, QString> NodeAttributeValues;
typedef QSet<Triple> TripleSet;

class RdfsGraph {
private:
    // Nodes (subjects/objects) and edges (RDFS Triples) of the graph
    QSet<QString> m_nodes;
    TripleSet     m_edges;

    // Attributes (markings) of nodes: map from node type to (node, attribute value) pair
    QMap<QString, NodeAttributeValues> m_nodeAttributes;
    // Attributes (markings) of edges: map from node type to (edge, attribute value) pair
    QMap<QString, EdgeAttributeValues> m_edgeAttributes;

    // Hash tables for incoming/outgoing triples of nodes
    QMap<QString, TripleSet> m_incomingEdges;
    QMap<QString, TripleSet> m_outgoingEdges;

public:
    /**
     * Constructor
     */
    RdfsGraph();

    /**
     * Reset the graph (remove all nodes and edges/triples)
     */
    void clear();

    /**
     * Add a list of triples to the graph
     * @param triples The list of triples to add
     */
    void add(QList<Triple> triples);

    /**
     * Add a set of triples to the graph.
     * @param triples The set of triples to add
     */
    void add(QSet<Triple> triples);

    /**
     * Add a single triple to the graph. If the triple already exists, no changes
     * will be made.
     * @param t The triple to add
     */
    void add(Triple t);

    /**
     * Remove a triple from the graph. If the triple does not exist, no changes
     * will be made. Note that the triple may still be derivable through RDFS
     * semantics after this operation.
     * @param t The triple to remove.
     */
    void remove(Triple t);

    /**
     * Assign an attribute to a node (subject or object of a triple). The node
     * must exist in the graph (as node), otherwise the attribute will not be
     * assigned. If a value was already set, the new value is added to a list:
     * the previously assigned value remains.
     * @param attribute the name of the attribute to add
     * @param node The name of the node to assign the attribute to
     * @param value The value to assign
     */
    void setNodeAttribute(QString attribute, QString node, QString value);
    /**
     * Get the values assigned to nodes for a specific attribute
     * @param attribute The attribute to retrieve the values of.
     */
    inline NodeAttributeValues nodeAttributes(QString attribute) const { return m_nodeAttributes.value(attribute); };
    /**
     * Get the value assigned to a specific node for a specific attribute
     * @param attribute The attribute to retrieve the value of.
     * @param node The node to get the attribute value of.
     */
    inline QString nodeAttribute(QString attribute, QString node) const { return m_nodeAttributes.value(attribute).value(node); };

    /**
     * Assign an attribute to a edge (triple). The edge must exist in the graph
     * otherwise the attribute will not be assigned. If a value was already set,
     * the new value is added to a list: the previously assigned value remains.
     * @param attribute the name of the attribute to add
     * @param edge The triple to assign the attribute to
     * @param value The value to assign
     */
    void setEdgeAttribute(QString attribute, Triple edge, QString value);
    /**
     * Get the values assigned to edges for a specific attribute
     * @param attribute The attribute to retrieve the values of.
     */
    inline EdgeAttributeValues edgeAttributes(QString attribute) const { return m_edgeAttributes.value(attribute); };
    /**
     * Get the value assigned to a specific edge for a specific attribute
     * @param attribute The attribute to retrieve the values of.
     * @param edge The edge to get the attribute value of.
     */
    inline QString edgeAttribute(QString attribute, Triple edge) const { return m_edgeAttributes.value(attribute).value(edge); };

    /**
     * Get all incoming edges (triples) of a specific node. All these triples will have
     * the same object, namely the given node.
     * @param node The node to get the incoming edges of.
     */
    inline TripleSet incomingEdges(QString node) const { return m_incomingEdges.value(node); };
    /**
     * Get all outgoing edges (triples) of a specific node. All these triples will have
     * the same subject, namely the given node.
     * @param node The node to get the outgoing edges of.
     */
    inline TripleSet outgoingEdges(QString node) const { return m_outgoingEdges.value(node); };

    /**
     * Get the set of all nodes in the graph
     */
    inline QSet<QString> nodes() const { return m_nodes; };
    /**
     * Get the set of all triples in the graph
     */
    inline TripleSet edges() const { return m_edges; };

    ~RdfsGraph();
};

}

/**
    * Add a single triple to a graph. If the triple already exists, no changes
    * will be made.
    * @param g The graph to add the triple to
    * @param t The triple to add
    */
Rdfs::RdfsGraph operator<< (Rdfs::RdfsGraph &g, const Rdfs::Triple &t);

#endif // RDFS_RDFSGRAPH_H
