/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "typerule.h"

#include <QQueue>
#include <QStack>
#include <QStringList>

#include "../rdfsgraph.h"
#include "../rdfslogger.h"
#include "../rdfscursor.h"

using namespace Rdfs;

TypeRule::TypeRule()
{

}

bool TypeRule::check(Rdfs::Triple t, Rdfs::RdfsTable table, Rdfs::RdfsBackend* store)
{
    bool tInEntailment = false;

    // Since we'll use these a number of times, don't do the lookup every time
    const QString sp = map(RDFS_SP);
    const QString sc = map(RDFS_SC);
    const QString type = map(RDFS_TYPE);
    const QString dom = map(RDFS_DOM);
    const QString range = map(RDFS_RANGE);

    const unsigned short maxScDistance = (((store->capabilities() & RDFS_TRANSITIVE) > 0) && store->transitivePredicates().contains(sc))
                                     ? store->transitiveDistance()
                                     : 0;
    const unsigned short maxSpDistance = (((store->capabilities() & RDFS_TRANSITIVE) > 0) && store->transitivePredicates().contains(sp))
                                     ? store->transitiveDistance()
                                     : 0;

    if (type.isEmpty() || t.predicate != type) {
        return false;
    }

    // Extra triples to be used throughout the algorithm
    Triple ta, tb;

    // For each triple (u,dom,v) in G, mark the vertex u in G(sp) with d(v)
    // See note below why this is skipped

    // Construct Ld: list of elements marked d(v) such that a path v->b exists in gsc
    // As per Munoz et al.: List all d(v) and expand gsc from t.object, each time marking
    // elements from the list. Remove unmarked elements at the end.
    //
    // NOTE: We do not list all d(v) first. Instead, for each item we find (every Y (X,sc,Y), starting
    // with Y=t.object), we check:
    // - (Z',dom,Y) exists -> add Z' to Ld
    // - (Z'',range,Y) exists -> Add Z'' to Lr (Step 6)
    //
    // Additionally we check if (t.subject, type, Y) for each Y (for which (Y, sc, t.object) holds). If so,
    // we have a positive match.
    QSet<QString> Ld;
    QSet<QString> Lr;
    RdfsCursor *c, *ca;
    QStack<RdfsCursor *> cursors;
    int depth = -1;
    QSet<QString> processed; // to prevent cycles
    QString current;

    RdfsLogger::instance()->debug("Initialized");

    // Check some extra triples for the first round. Note that if no (X,sc,t.object) exists, we still
    // want to find results if (Y,dom,t.object) does exist, or example. However, opening the cursor
    // for (X,sc,t.object) may fail!
    if (!dom.isEmpty()) {
        ta.subject = QString();
        ta.predicate = dom;
        ta.object = t.object;
        c = store->cursor(ta, table);
        if (c != 0 && c->open()) {
            while (!c->eof()) {
                tb = c->current();
                Ld << tb.subject;
                c->next();
            }
            c->close();
            delete c;
        }
        else if (c != 0) {
            delete c;
        }
    }
    // range
    if (!range.isEmpty()) {
        ta.subject = QString();
        ta.predicate = range;
        ta.object = t.object;
        c = store->cursor(ta, table);
        if (c != 0 && c->open()) {
            while (!c->eof()) {
                tb = c->current();
                Lr << tb.subject;
                c->next();
            }
            c->close();
            delete c;
        }
        else if (c != 0) {
            delete c;
        }
    }

    RdfsLogger::instance()->debug("Initial Ld/Lr built");

    // Now start initial cursor
    ta.subject = QString();
    ta.predicate = sc;
    ta.object = t.object;
    c = store->cursor(ta, table);
    if (c != 0 && c->open()) {
        cursors.push(c);
        depth = 0;
    }
    else if (c != 0) {
        delete c;
    }
    processed << t.object;

    while (depth >= 0) {
        // Note: check subject of each triple here, otherwise we'll miss the first in the chain
        ta = c->current();

        RdfsLogger::instance()->debug(QString("Building Ld/Lr: Processing ta = (%1, %2, %3)").arg(ta.subject, ta.predicate, ta.object));

        // Now traverse over subclasses
        // From ta = (X,sc,Y) -> (Z,sc,X)
        if (!processed.contains(ta.subject)) {
            processed << ta.subject;

            if (!dom.isEmpty()) {
                // Check (D, dom, ta.subject)
                tb.subject = QString();
                tb.predicate = dom;
                tb.object = ta.subject;
                ca = store->cursor(tb, table);
                if (ca != 0 && ca->open()) {
                    while (!ca->eof()) {
                        // Add D to Ld
                        tb = ca->current();
                        Ld << tb.subject;
                        ca->next();
                    }
                    ca->close();
                    delete ca;
                }
                else if (ca != 0) {
                    delete ca;
                }
            }

            // range
            if (!range.isEmpty()) {
                // Check (R, range, ta.subject)
                tb.subject = QString();
                tb.predicate = range;
                tb.object = ta.subject;
                ca = store->cursor(tb, table);
                if (ca != 0 && ca->open()) {
                    while (!ca->eof()) {
                        // Add R to Lr
                        tb = ca->current();
                        Lr << tb.subject;
                        ca->next();
                    }
                    ca->close();
                    delete ca;
                }
                else if (ca != 0) {
                    delete ca;
                }
            }

            // type ("type" is never empty, as it's also in the query)
            // Note that we can directly query (t.subject, type, ta.subject). Note that
            // (ta.subject, sc, t.object), and as such if the triple is found,
            // we also know (t.subject type t.object) (i.e. the query)
            tb.subject = t.subject;
            tb.predicate = type;
            tb.object = ta.subject;
            if (store->hasTriple(tb, table)) {
                // Let while() loop below clean up remaining cursors
                tInEntailment = true;
            }

            // Build next triple, if distance is at max sc distance
            if (ta.distance >= maxScDistance) {
                RdfsLogger::instance()->debug(QString("Traversing for search = (?, %2, %3)").arg(sc, ta.subject));

                tb.object = ta.subject;
                tb.subject = QString();
                tb.predicate = sc;

                ca = store->cursor(tb, table);
                if (ca != 0 && ca->open()) {
                    // Move into depth
                    cursors.push(c); // Push parent onto stack
                    c = ca;
                    ++depth;
                    continue; // Immediately use this cursor
                }
                else if (ca != 0) {
                    delete ca;
                }
            }
        }

        // Move to next item
        c->next();

        // If we processed the rightmost triple in tree: move to parent
        while (c->eof() || tInEntailment) {
            c->close(); // Clean up current
            delete c;
            --depth;
            if (depth >= 0) {
                RdfsLogger::instance()->debug(QString("Moving up (to depth: %1)...").arg(QString::number(depth)));
                c = cursors.pop();
                c->next();
            }
            else { // depth < 0 will stop main loop - i.e. we're done
                cursors.pop(); // Do not use return value
                break; // No valid reference 'c' anymore
            }
        } // while c->eof()
    } // Built Ld, Lr, Lt
    processed.clear();

    // If we found a triple (because of (a,type,y) and (y,sc,b)), exit here
    if (tInEntailment) {
        return true;
    }

    RdfsLogger::instance()->debug("Built Ld/Lr...");

    RdfsLogger::instance()->debug(QString("Ld: {%1}").arg(QStringList(Ld.toList()).join(", ")));
    RdfsLogger::instance()->debug(QString("Lr: {%1}").arg(QStringList(Lr.toList()).join(", ")));

    // Note: Ld, Lr now contain all property nodes of which the domain or range (resp.) is a superclass
    // of t.object

    // Construct Ls: ordered list of elements s(z) such that:
    //               1) gsp has path from node marked s(z) to a node marked
    //                  with an element in Ld
    //               2) there is (z, type, v) in G, for d(v) in Ld

    // Note: rather than constructing Ls, we can terminate as soon as we find
    // the right property. Since we have not explicitly marked all s(z) items
    // earlier, we take a slightly different approach to reach the same goal:
    // - Enqueue ("mark green" in Munoz et al description) all nodes in Ld + Lr
    // - While queue is not empty:
    //   - Dequeue p
    //   - If p or ancestor in Ld: check if (a, p, x) exists for any x
    //   - If p or ancestorin Lr: check if (x, p, a) exists for any x
    //   - Enqueue all q for each triple (q, sp, p)
    // - If queue is empty without positive check: (a, type, b) is not in entailment
    //
    // Implementation detail: to prevent going into a sp-tree's breadth, we take the
    // DFS approach similar to the sc-traversal above, by making a stack of cursors.
    // Note that the general behavior of sp/sc is tree-like, which goes more into depth
    // than breadth. This requires at most the depth of the tree of memory space (times
    // the size of the cursor plus a pointer to that cursor and some constant overhead).
    //
    // We keep an extra list of processed properties (those marked green) in memory,
    // to prevent endless loops due to cycles.

    QQueue<QString> queue;
    QStack<bool> inLd, inLr; // Info about parent ("parent or ancestor in Ld/Lr")
    queue.append((Ld + Lr).toList());

    // Note; in some cases, a property on its own may only be in Ld, although an
    // ancestor is in Lr. In this case, the property should be considered as being in
    // both Ld and Lr. We do not know this when we begin, and thus initially may only
    // traverse it for the "domain" (ignoring triples where it's used as range).
    //
    // There are two solutions around this:
    // - Track separately for domain/range which properties were processed. If a property
    //   was already processed in one of them and needs to be traversed again for the
    //   other, then enqueue it again for the other part. This may result in the subtree
    //   being traversed twice.
    //
    // - Test for both domain and range matches, regardless of the origin of the property.
    //   If the property (or ancestor) is not in Ld but there is a match on domain, then
    //   temporarily store the property (and all parents) in a special list.
    //   If later you need to revisit it (e.g. because of a domain match), it suffices to check
    //   if the triple occurs in the list. The drawback is that more triples are (unnecessarily)
    //   checked all the time, for the entire sp-subtree. You don't have to walk a subtree
    //   twice, though.
    //
    // The solution implemented here is the first. It is less memory-intensive. Also, note that
    // in the proposed algorithm the subtree may be walked twice as well, considering Step 5 and 6
    // are separated and the match does not exist or is only because of the range; the
    // subproperties would still be traversed for the domain, and afterwards again for the range.
    // Hence it is not less efficient than the original algorithm, at least when it comes to
    // look-ups in the database.

    QSet<QString> processedDomain, processedRange;

    QString p;
    while (!queue.isEmpty() && !tInEntailment) {
        p = queue.dequeue();

        // If we already processed p for the domain/range resp., do not
        // reconsider that again
        inLd.push(Ld.contains(p) && !processedDomain.contains(p));
        inLr.push(Lr.contains(p) && !processedRange.contains(p));

        // Check if we need to process this triple again
        if (!inLd.top() && !inLr.top()) {
            // Skip this one, already processed it before for the required
            // domain or range items
            continue;
        }

        RdfsLogger::instance()->debug(QString("Processing property: %1").arg(p));

        // Do DFS for p
        // First check p itself
        if (inLd.top()) {
            // Mark as processed
            processedDomain << p;
            tb.subject = t.subject;
            tb.predicate = p;
            tb.object = QString();

            if (store->hasTriple(tb, table)) {
                // Result found (a, p, x) with (p, dom, y) and (y, sc, b)
                tInEntailment = true; // Set to true and allow clean-up to be done
                break;
            }
        }

        if (inLr.top()) {
            processedRange << p;
            tb.subject = QString();
            tb.predicate = p;
            tb.object = t.subject;

            if (store->hasTriple(tb, table)) {
                // Result found (x, p, a) with (p, range, y) and (y, sc, b)
                tInEntailment = true; // Set to true and allow clean-up to be done
                break;
            }
        }

        // Find initial (q, sp, p) triples
        depth = -1;
        ta.subject = QString();
        ta.predicate = sp;
        ta.object = p;
        c = store->cursor(ta, table);
        if (c != 0 && c->open()) {
            depth = 0;
            cursors.push(c);
        }
        else if (c != 0) {
            delete c;
        }

        while (depth >= 0) {
            ta = c->current();

            // Check for which subtrees (domain/range) we need to check this
            inLd.push((inLd.top() || Ld.contains(ta.subject)) && !processedDomain.contains(ta.subject));
            inLr.push((inLr.top() || Lr.contains(ta.subject)) && !processedRange.contains(ta.subject));

            // Either inLd.top() or inLr.top() is true if:
            // - The property (ta.subject) or an ancestor appears in Ld (domain) and/or Lr (range) resp.
            // - The property was not yet traversed for this appearance
            if (inLd.top() || inLr.top()) {
                // Process item
                RdfsLogger::instance()->debug(QString("Processing subproperty: %1").arg(ta.subject));

                if (inLd.top()) {
                    processedDomain << ta.subject;
                    tb.subject = t.subject;
                    tb.predicate = ta.subject;
                    tb.object = QString();
                    if (store->hasTriple(tb, table)) {
                        // Result found (a, q, x) with (q, sp, p), (p, dom, y) and (y, sc, b)
                        RdfsLogger::instance()->debug(QString("Found (%1, %2, _)").arg(t.subject, ta.subject));
                        tInEntailment = true; // Set to true and allow clean-up to be done
                    }
                    else RdfsLogger::instance()->debug(QString("NOT Found (%1, %2, _)").arg(t.subject, ta.subject));
                }

                if (inLr.top()) {
                    processedRange << ta.subject;
                    tb.subject = QString();
                    tb.predicate = ta.subject;
                    tb.object = t.subject;
                    if (store->hasTriple(tb, table)) {
                        // Result found (x, q, a) with (q, sp, p), (p, range, y) and (y, sc, b)
                        RdfsLogger::instance()->debug(QString("Found (_, %1, %2)").arg(ta.subject, t.subject));
                        tInEntailment = true; // Set to true and allow clean-up to be done
                    }
                    else RdfsLogger::instance()->debug(QString("NOT Found (_, %1, %2)").arg(ta.subject, t.subject));
                }

                // Go further into sp-tree if we have no result yet
                if (!tInEntailment && ta.distance >= maxSpDistance) {
                    tb.subject = QString();
                    tb.predicate = sp;
                    tb.object = ta.subject;
                    ca = store->cursor(tb, table);
                    if (ca != 0 && ca->open()) {
                        // Move into depth
                        cursors.push(c); // Push parent
                        ++depth;
                        c = ca;
                        continue; // Use cursor immediately, do not skip to 2nd position!
                    }
                    else if (ca != 0) {
                        delete ca;
                    }
                }
            }

            // Top inLd/inLr items are of current triple. Note that they're popped
            // when going up in the stack, if the current triple has children.
            inLd.pop();
            inLr.pop();

            // Advance cursor, if we're at the right (or found a match), move up
            c->next();
            while (c->eof() || tInEntailment) {
                // Go to parent cursor
                c->close();
                delete c;
                depth--;
                // Also pop inLd/inLr of "new" current cursor. If depth < 0, this will pop
                // the initial values (resulting in empty stacks)
                inLd.pop();
                inLr.pop();
                if (depth >= 0) {
                    c = cursors.pop();
                    c->next();
                }
                else { // depth < 0 will stop main loop - i.e. we're done
                    cursors.pop(); // Do not use return value
                    break; // No valid reference 'c' anymore
                }
            }
        } // while depth >= 0
    } // while !queue.isEmpty()

    // tInEntailment was set to true if a match was found, which will have stopped the loops.
    // If we search all the trees without a match, tInEntailment will still be false.
    return tInEntailment;
}


void TypeRule::setAttribute(QString node, QString attribute, QString value)
{
    Triple t;
    t.subject = node;
    t.predicate = "rdfsupdate:attribute:" + attribute;
    t.object = value;
    m_store->insertTriple(t, m_table);
}

RdfsCursor* TypeRule::predicateCursor(QString predicate)
{
    Triple t;
    t.subject = QString();
    t.predicate = predicate;
    t.object = QString();
    return m_store->cursor(t, m_table);
}

RdfsCursor* TypeRule::attributeCursor(QString attribute)
{
    return predicateCursor("rdfsupdate:attribute:" + attribute);
}

bool TypeRule::haveNode(QString node, QString predicate)
{
    Triple t;
    t.subject = node;
    t.predicate = predicate;
    t.object = QString();
    if (m_store->hasTriple(t, m_table)) {
        return true;
    }
    qSwap<QString>(t.subject, t.object);
    return m_store->hasTriple(t, m_table);
}


TypeRule::~TypeRule()
{

}

