/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "rdfssoprano.h"

#include <QList>
#include <Soprano/QueryResultIterator>
#include <Soprano/StatementIterator>
#include <Soprano/Inference/Rule>
#include "rdfslogger.h"

using namespace Rdfs;
using namespace Soprano;

RdfsSoprano::RdfsSoprano(Soprano::Model* model)
    : RdfsBackend(), m_model(model)
{
}

RdfsSoprano::~RdfsSoprano()
{
}

bool RdfsSoprano::isOpen()
{
    return (m_model != 0);
}

void RdfsSoprano::close()
{
    delete m_model;
    m_model = 0;
}

int RdfsSoprano::capabilities()
{
    return RDFS_NONE;
}


void RdfsSoprano::insertTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table)
{
    const Soprano::Node context = tableContext(table);
    Statement s = makeStatement(t);
    s.setContext(context);
    model()->addStatement(s);
}

int RdfsSoprano::checkTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table)
{
    const Soprano::Node context = tableContext(table);
    Statement s = makeStatement(t);
    s.setContext(context);
    return model()->listStatements(s).allStatements().size();
}

QList<Rdfs::Triple> RdfsSoprano::findTriples(const Rdfs::Triple& t, Rdfs::RdfsTable table)
{
    const Soprano::Node context = tableContext(table);
    Statement s = makeStatement(t);
    s.setContext(context);
    StatementIterator i = model()->listStatements(s);
    QList<Rdfs::Triple> result;
    Q_FOREACH(Statement ss, i.allStatements()) {
        result << makeTriple(ss);
    }
    return result;
}

void RdfsSoprano::removeTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table)
{
    const Soprano::Node context = tableContext(table);
    Statement s = makeStatement(t);
    s.setContext(context);

    model()->removeAllStatements(s);
}

void RdfsSoprano::insertRule(Rule r)
{
    const Soprano::Node context = tableContext(RDFS_RULE);
    Statement s;
    s.setSubject(QUrl(serializeTriple(r.first)));
    s.setPredicate(QUrl("rdfsupdate://rule/infers"));
    s.setObject(QUrl(serializeTriple(r.second)));
    s.setContext(context);

    model()->addStatement(s);
}

void RdfsSoprano::removeRule(Rule r)
{
    const Soprano::Node context = tableContext(RDFS_RULE);
    Statement s;
    s.setSubject(QUrl(serializeTriple(r.first)));
    s.setPredicate(QUrl("rdfsupdate://rule/infers"));
    s.setObject(QUrl(serializeTriple(r.second)));
    s.setContext(context);

    model()->removeAllStatements(s);
}

QList< Rdfs::Rule > RdfsSoprano::rules()
{
    const Soprano::Node context = tableContext(RDFS_RULE);
    StatementIterator i = model()->listStatementsInContext(context);
    QList<Statement> allStatements = i.allStatements();
    QList<Rdfs::Rule> result;
    Q_FOREACH(Statement s, allStatements) {
        Rdfs::Rule r;
        r.first = unserializeTriple(s.subject().toString());
        r.second = unserializeTriple(s.object().toString());
        if (s.predicate().toString() == "rdfsupdate://rule/infers") {
            result << r;
        }
    }
    return result;
}

void RdfsSoprano::truncate(Rdfs::RdfsTable table)
{
    const Soprano::Node context = tableContext(table);
    model()->removeContext(context);
}
