/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef RDFS_RDFSSOPRANO_H
#define RDFS_RDFSSOPRANO_H

#include <QStringList>

#include <Soprano/Model>
#include <Soprano/Inference/InferenceModel>
#include <Soprano/Node>
#include <Soprano/Statement>
#include <Soprano/Vocabulary/RDF>
#include <Soprano/Vocabulary/RDFS>

#include "rdfsbackend.h"

#include "rdfssql.h"
#include "rdfslogger.h"

namespace Rdfs {

class RdfsSoprano : public RdfsBackend
{
private:
    Soprano::Model *m_model;

    /**
     * Converts a string to a Soprano Node. Also converts known RDFS names and abbreviations
     * using the proper namespace.
     */
    Soprano::Node stringToNode(const QString &literal);

    /**
     * Serialize a triple to a single QString, which can be reused as single literal
     * in new triples.
     * @param t The triple to serialize
     * @return The serializes triple
     */
    QString serializeTriple(Triple t);

    /**
     * Unserialize a serialized triple.
     * @param t The triple to unserialize, as QString from serializeTriple()
     * @return The unserialized Triple
     */
    Triple unserializeTriple(QString t);

    /**
     * Create a context node from a table name
     * @param table The table to create the context node of.
     */
    Soprano::Node tableContext(RdfsTable table);

protected:
    explicit RdfsSoprano(Soprano::Model *model = 0);
    Soprano::Model *model();
    void setModel(Soprano::Model *model);

    Soprano::Statement makeStatement(const Rdfs::Triple &t);

    Triple makeTriple(const Soprano::Statement &s);

public:
    virtual ~RdfsSoprano();

    virtual bool isOpen();
    virtual void close();

    virtual int capabilities();

    virtual void insertTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table = RDFS_ORIGINAL);
    virtual int checkTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table = RDFS_ORIGINAL);
    virtual QList< Triple > findTriples(const Rdfs::Triple& t, RdfsTable table);
    virtual void removeTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table = RDFS_ORIGINAL);

    virtual void insertRule(Rule r);
    virtual void removeRule(Rule r);
    virtual QList< Rdfs::Rule > rules();
    virtual void truncate(Rdfs::RdfsTable table = RDFS_ORIGINAL);
};

// -----------------------------------------------------------------------------
inline Soprano::Model* RdfsSoprano::model()
{
    return m_model;
}

inline void RdfsSoprano::setModel(Soprano::Model* model)
{
    m_model = model;
}

inline QString RdfsSoprano::serializeTriple(Triple t)
{
    QString result = QString::number(t.subject.length()) + ";" +
                     QString::number(t.predicate.length()) + ";" +
                     QString::number(t.object.length()) + ";" +
                     t.subject + ";" + t.predicate + ";" + t.object;
    return result;
}

inline Triple RdfsSoprano::unserializeTriple(QString t)
{
    Triple result;
    int subjectLength, predicateLength, objectLength;
    QStringList data = t.split(";", QString::KeepEmptyParts);
    if (data.size() < 6) {
        return result;
    }

    bool ok;
    subjectLength = data.at(0).toInt(&ok);
    if (!ok) { return result; }
    predicateLength = data.at(1).toInt(&ok);
    if (!ok) { return result; }
    objectLength = data.at(2).toInt(&ok);
    if (!ok) { return result; }

    int offset = data.at(0).size() + data.at(1).size() + data.at(2).size() + 3;

    // Invalid length of total string?
    if (t.size() != offset + subjectLength + predicateLength + objectLength + 2) {
        return result;
    }

    result.subject = t.mid(offset, subjectLength);
    result.predicate = t.mid(offset + subjectLength + 1, predicateLength);
    result.object = t.mid(offset + subjectLength + predicateLength + 2, objectLength);

    return result;
}

inline Soprano::Statement RdfsSoprano::makeStatement(const Rdfs::Triple& t)
{
    return Soprano::Statement(stringToNode(t.subject),
                              stringToNode(t.predicate),
                              stringToNode(t.object));
}

inline Triple RdfsSoprano::makeTriple(const Soprano::Statement& s)
{
    Triple t;
    t.subject = s.subject().toString();
    t.predicate = s.predicate().toString();
    t.object = s.object().toString();
    return t;
}


inline Soprano::Node RdfsSoprano::stringToNode(const QString& literal)
{
    QUrl url(literal, QUrl::StrictMode);

    // Check empty literal
    if (literal.isEmpty()) {
        return Soprano::Node::createEmptyNode();
    }
    // Check RDF/RDFS literals + abbreviations
    else if (literal == "sp" || literal == "subPropertyOf" || literal == "rdfs:subPropertyOf") {
        return Soprano::Vocabulary::RDFS::subPropertyOf();
    }
    else if (literal == "sc" || literal == "subClassOf" || literal == "rdfs:subClassOf") {
        return Soprano::Vocabulary::RDFS::subClassOf();
    }
    else if (literal == "dom" || literal == "domain" || literal == "rdfs:domain") {
        return Soprano::Vocabulary::RDFS::domain();
    }
    else if (literal == "range" || literal == "range" || literal == "rdfs:range") {
        return Soprano::Vocabulary::RDFS::range();
    }
    else if (literal == "type" || literal == "type" || literal == "rdf:type") {
        return Soprano::Vocabulary::RDF::type();
    }

    else if (literal == "Class" || literal == "class" || literal == "rdfs:Class") {
        return Soprano::Vocabulary::RDFS::Class();
    }
    else if (literal == "Property" || literal == "property" || literal == "rdf:Property") {
        return Soprano::Vocabulary::RDF::Property();
    }

    else if (literal == "Container" || literal == "rdfs:Container") {
        return Soprano::Vocabulary::RDFS::Container();
    }
    else if (literal == "ContainerMembershipProperty" || literal == "rdfs:ContainerMembershipProperty") {
        return Soprano::Vocabulary::RDFS::ContainerMembershipProperty();
    }
    else if (literal == "Datatype" || literal == "rdfs:Datatype") {
        return Soprano::Vocabulary::RDFS::Datatype();
    }
    else if (literal == "Literal" || literal == "rdfs:Literal") {
        return Soprano::Vocabulary::RDFS::Literal();
    }
    else if (literal == "Resource" || literal == "rdfs:Resource") {
        return Soprano::Vocabulary::RDFS::Resource();
    }
    else if (literal == "comment" ||literal == "rdfs:comment") {
        return Soprano::Vocabulary::RDFS::comment();
    }
    else if (literal == "isDefinedBy " || literal == "rdfs:isDefinedBy ") {
        return Soprano::Vocabulary::RDFS::isDefinedBy();
    }
    else if (literal == "label" || literal == "rdfs:label") {
        return Soprano::Vocabulary::RDFS::label();
    }
    else if (literal == "member" || literal == "rdfs:member") {
        return Soprano::Vocabulary::RDFS::member() ;
    }
    else if (literal == "seeAlso" || literal == "rdfs:seeAlso") {
        return Soprano::Vocabulary::RDFS::seeAlso() ;
    }

    else if (literal == "Alt" || literal == "rdf:Alt") {
        return Soprano::Vocabulary::RDF::Alt();
    }
    else if (literal == "Bag" || literal == "rdf:Bag") {
        return Soprano::Vocabulary::RDF::Bag();
    }
    else if (literal == "List" || literal == "rdf:List") {
        return Soprano::Vocabulary::RDF::List();
    }
    else if (literal == "Seq" || literal == "rdf:Seq") {
        return Soprano::Vocabulary::RDF::Seq();
    }
    else if (literal == "Statement" || literal == "rdf:Statement") {
        return Soprano::Vocabulary::RDF::Statement();
    }
    else if (literal == "XMLLiteral" || literal == "rdf:XMLLiteral") {
        return Soprano::Vocabulary::RDF::XMLLiteral();
    }
    else if (literal == "first" || literal == "rdf:first") {
        return Soprano::Vocabulary::RDF::first();
    }
    else if (literal == "nil" || literal == "rdf:nil") {
        return Soprano::Vocabulary::RDF::nil();
    }
    else if (literal == "object" || literal == "rdf:object") {
        return Soprano::Vocabulary::RDF::object();
    }
    else if (literal == "predicate" || literal == "rdf:predicate") {
        return Soprano::Vocabulary::RDF::predicate();
    }
    else if (literal == "rest" || literal == "rdf:rest") {
        return Soprano::Vocabulary::RDF::rest();
    }
    else if (literal == "subject" || literal == "rdf:subject") {
        return Soprano::Vocabulary::RDF::subject();
    }
    else if (literal == "value" || literal == "rdf:value") {
        return Soprano::Vocabulary::RDF::value();
    }

    // No match: return literal as rdfsupdate resource node
    else if (url.isValid()) {
        return Soprano::Node(url);
    }
    else {
        return Soprano::Node(QUrl("http://rdfsupdate/literal/" + literal));
    }
}

inline Soprano::Node RdfsSoprano::tableContext(RdfsTable table)
{
    switch(table) {
        case RDFS_ORIGINAL:
        default:
            return Soprano::Node(QUrl("http://rdfsupdate/tables/original"));
        case RDFS_UPDATE:
            return Soprano::Node(QUrl("http://rdfsupdate/tables/update"));
        case RDFS_RULE:
            return Soprano::Node(QUrl("http://rdfsupdate/tables/rule"));
    }
    return Soprano::Node(QUrl("http://rdfsupdate/tables/original"));
}
}

#endif // RDFS_RDFSSOPRANO_H
