/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "rdfstripletbackend.h"

#include "rdfslogger.h"
#include "rdfstripletcursor.h"
#include <QSettings>

using namespace Rdfs;

const quint64 RdfsTripleTBackend::m_maxBulkLoadSize = 100000;
const uint RdfsTripleTBackend::m_maxPrimaryPayloadPages = 4;

const QStringList RdfsTripleTBackend::m_transitivePredicates = QStringList() << "sp" << "sc";

RdfsTripleTBackend::RdfsTripleTBackend(QString settingsPath)
{
    m_inBulkLoad = false;
    m_bulkLoadCount = 0;
    m_maxPrimaryPayloadSize = 4096 * m_maxPrimaryPayloadPages; // 4 pages of 4Kb
#ifndef NDEBUG
    m_savingStatistics = false;
    m_loadedStatistics = false;
#endif
    m_settingsPath = settingsPath;

    // Load settings
    QSettings s;
    m_distance = (unsigned char)(s.value(settingsPath + "/distance", 0).toUInt());
}

int RdfsTripleTBackend::capabilities() const
{
    return RDFS_TRANSITIVE;
}

bool RdfsTripleTBackend::setOption(const QString& key, const QString& value)
{
    if (key == "distance") {
        bool ok;
        unsigned char distance = (unsigned char)(value.toUInt(&ok));
        if (ok && distance < 128) { // Distance of 128 or more will cause overflows!
            m_distance = distance;
            QSettings s;
            s.setValue(m_settingsPath + "/distance", distance);
            return true;
        }
        else {
            RdfsLogger::instance()->error("Distance setting must be a number between 0 and 127");
        }
    }
    return false;
}

unsigned char RdfsTripleTBackend::transitiveDistance() const
{
    return m_distance;
}

QStringList RdfsTripleTBackend::transitivePredicates() const
{
    return m_transitivePredicates;
}


void RdfsTripleTBackend::setPageSize(uint size)
{
    if (size > 4) {
        m_maxPrimaryPayloadSize = size * m_maxPrimaryPayloadPages;
    }
}

QList< TertiaryAtom >::iterator RdfsTripleTBackend::binaryFindTertiary(QList< TertiaryAtom >* list, const QString& tertiary)
{
    QList<TertiaryAtom>::iterator i, j, mid;
    i = list->begin();
    j = list->end();

    while (j - i > 0) { // Note: "i < j" would compare the values, not the distance of the iterators!
        mid = i + ((j - i) / 2);
        if ((*mid).first < tertiary) {
            i = mid + 1;
        }
        else {
            j = mid;
        }
    }

    // If item already exists, it will now be at position i
    if (i != list->end() && (*i).first == tertiary) {
        return i;
    }
    return list->end();
}


void RdfsTripleTBackend::removeTriple(const Rdfs::Triple& t, RdfsTable table)
{
    if (t.subject.isEmpty() && t.predicate.isEmpty() && t.object.isEmpty()) {
        truncate();
        return;
    }

    // Load all triples prior to removing them. Cursors could potentially go corrupt,
    // if data is removed while the cursor is still open.
    // Note: even with complete triple, find it to ensure we have the correct distance info
    QList<Triple> triples = findTriples(t, table);

    // Note: if t.distance >= m_distance, there won't be inferred triples, as those would
    // have distance > m_distance - i.e., no need to do inference
    bool inference = (isTransitive(t.predicate) && t.distance < m_distance);
    Q_FOREACH(Triple tt, triples) {
        removeCompleteTriple(tt, table);
        if (inference) {
            removeInferredTriple(tt, table);
        }
    }
}

void RdfsTripleTBackend::removeCompleteTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table)
{
    RdfsLogger::instance()->debug(QObject::tr("**** Removing: %1 %2 %3").arg(t.subject, t.predicate, t.object));
    // Code re-use: create interim "primary, secondary, tertiary" triples for index
    Triple tsop, tpso, tosp;
    tsop.subject = tpso.predicate = tosp.predicate = t.subject;
    tsop.object = tpso.subject = tosp.object = t.predicate;
    tsop.predicate = tpso.object = tosp.subject = t.object;
    QMap<RdfsRole, Triple> roles;
    QMap<RdfsRole, Triple>::const_iterator i;
    roles.insert(RDFS_SUBJECT, tsop);
    roles.insert(RDFS_PREDICATE, tpso);
    roles.insert(RDFS_OBJECT, tosp);

    for (i = roles.constBegin(); i != roles.constEnd(); ++i) {
        RdfsRole role = i.key();
        Triple tt = i.value();

        // Load data of primary for given role
        TripleTPayload data = deserializePayload(payload(tt.subject, role, table));
        if (!data.hasData) {
            removeBtreeTriple(tt.subject, role, tt.predicate, tt.object, table);
            continue;
        }

        QList<TertiaryAtom> tertiaryList;
        QList<TertiaryAtom>::iterator iTertiary;
        if (!data.data.contains(tt.predicate)) { // Secondary
            // Secondary doesn't even exist, i.e. triples doesn't exist -- nothing to remove
            return;
        }
        tertiaryList = data.data.value(tt.predicate);
        iTertiary = binaryFindTertiary(&tertiaryList, tt.object);
        if (iTertiary == tertiaryList.end()) {
            // Triple doesn't exist in this index (tertiary not found) -- nothing to remove
            return;
        }

        tertiaryList.erase(iTertiary); // Remove tertiary
        if (tertiaryList.size() > 0) {
            data.data.insert(tt.predicate, tertiaryList); // Update secondary list
        }
        else {
            data.data.remove(tt.predicate); // No triples "A B ?" anymore -> no need to store empty sublist
        }

        // Store update
        putPayload(tt.subject, role, serializePayload(data), table);
#ifndef NDEBUG
        recomputeStatistics(tt.subject, role, data);
#endif
    }
}

void RdfsTripleTBackend::removeInferredTriple(const Rdfs::Triple& t, RdfsTable table)
{
    QList<Triple> inferred = findInferredTriples(t, table);
    QList<Triple> removed;
    removed << t; // Assume t was removed before calling this method

    // Check which of the inferred triples are really materialized (by checking distance)
    Q_FOREACH(Triple it, inferred) {
        QList<Triple> result = findTriples(it, table);
        // Note: result should have exactly 1 result with smallest distance of inference -
        // i.e. it.distance or something smaller
        if (!result.isEmpty() && result.at(0).distance == it.distance) {
            removed << it;
            removeCompleteTriple(it, table);
        }
    }

    // Table is now clear; now re-add still-inferred triples (in bulk), including t if t is
    // still inferred.
    //
    // Assume t == (a,p,b). If inferrable within bounds, there still exists a
    // t2=(a,p,C)@0 and t3=(C,p,b)@x, with x <= max distance. Note that x+1 should
    // also be >= t.distance, otherwise t would not be removed as it wouldn't have
    // been materialized
    //
    // I.e., in steps:
    // 1. Find all (a,p,C) triples with distance 0
    // 2. Check if (C,p,b) exists (possibly inferred through other route)
    // 3. If so and distance is ok: re-add triple, including its own
    //    inferred triples. Note that if a previous triple was not found
    //    as inferrable, but is inferrable through (a,p,b), it will now be
    //    added after all, keeping the one with the smallest distance in the
    //    end.

    Q_FOREACH(Triple rt, removed) {
        Triple search;
        unsigned char insertDistance = m_distance + 1;

        // Find
        search.subject = rt.subject;
        search.predicate = rt.predicate;
        search.object = QString();

        // Use cursor
        RdfsCursor *c = cursor(search, table);
        if (c->open()) {
            while (!c->eof()) {
                // t2 == (a,p,C)
                Triple t2 = c->current();
                if (t2.distance != 0) {
                    // Not a direct neighbor
                    c->next();
                    continue;
                }

                // Find (C,p,b) (t3) - one result, so no cursor
                Triple searchT3;
                searchT3.subject = t2.object;
                searchT3.predicate = t2.predicate;
                searchT3.object = rt.object;
                QList<Triple> t3 = findTriples(searchT3, table);

                if (!t3.isEmpty() && t3.at(0).distance + 1 < insertDistance) {
                    // Note: continue search, in case later a shorter path is found
                    // Also: insert after closing the cursor, so we don't break it
                    // in any way due to new triples
                    insertDistance = t3.at(0).distance + 1;
                }
                c->next();
            }
            c->close();
        }
        delete c;

        if (insertDistance <= m_distance) {
            rt.distance = insertDistance;
            insertTriple(rt, table);
        }
    }
}


QList< Rdfs::Triple > RdfsTripleTBackend::findTriples(const Rdfs::Triple& t, Rdfs::RdfsTable table)
{
    QList< Rdfs::Triple > result;

    RdfsCursor *c = cursor(t, table);
    if (c->open()) {
        while (!c->eof()) {
            result.append(c->current());
            c->next();
        }
        c->close();
    }
    delete c;

    return result;
}

int RdfsTripleTBackend::checkTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table)
{
    int result = 0;

    RdfsCursor *c = cursor(t, table);
    if (c->open()) {
        while (!c->eof()) {
            result++;
            c->next();
        }
        c->close();
    }
    delete c;

    return result;
}

QByteArray RdfsTripleTBackend::serializePayload(TripleTPayload payload)
{
    if (payload.hasData && payload.data.isEmpty()) {
        return QByteArray();
    }
    else if (!payload.hasData) {
        RdfsLogger::instance()->debug("WRITING no data payload");
        return QByteArray(2, 0xFF); // 0xFFFF
    }

    QByteArray result;
    result.reserve(payloadSize(payload));
    QDataStream stream(&result, QIODevice::ReadWrite);

    // Write the payload size
    stream << (quint32)payload.data.size();

    // Loop over map
    QMap<QString, QList<TertiaryAtom> >::const_iterator i;
    for (i = payload.data.constBegin(); i != payload.data.constEnd(); ++i) {
        // Write secondary key
        stream << i.key().toUtf8();
        stream << (quint32)i.value().size();
        // And tertiary items
        TertiaryAtom atom;
        Q_FOREACH(atom, i.value()) {
            stream << atom.first.toUtf8(); // Note: writes size + data
            stream << atom.second;
        }
    }

    return qCompress(result);
}

TripleTPayload RdfsTripleTBackend::deserializePayload(const QByteArray& payload)
{
    TripleTPayload result;
    result.hasData = true;

    if (payload.isEmpty()) {
        RdfsLogger::instance()->debug("Empty payload");
        return result;
    }
    else if (payload.size() == 2 && payload.at(0) == char(0xFF) && payload.at(1) == char(0xFF)) {
        result.hasData = false;
        RdfsLogger::instance()->debug("No data payload");
        return result;
    }

    QDataStream stream(qUncompress(payload));
    quint32 size;
    stream >> size;

    for (quint32 i = 0; i < size; ++i) {
        // Read secondary item
        QByteArray utf8buffer;
        QString secondary;
        QList<TertiaryAtom> tertiary;
        stream >> utf8buffer;
        secondary = QString::fromUtf8(utf8buffer.constData(), utf8buffer.size());;

        // Read item count
        quint32 subsize;
        stream >> subsize;
        for (quint32 j = 0; j < subsize; ++j) {
            // Read string for tertiary items
            stream >> utf8buffer;
            TertiaryAtom atom;
            atom.first = QString::fromUtf8(utf8buffer.constData(), utf8buffer.size());
            stream >> atom.second;
            tertiary << atom;
        }
        result.data.insert(secondary, tertiary);
    }
    return result;
}


void RdfsTripleTBackend::insertTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table)
{
    if (m_inBulkLoad) {
        // This will also add necessary inferred triples
        insertBulkTriple(t, table);
    }
    else {
        insertNonbulkTriple(t, table);
    }
}

void Rdfs::RdfsTripleTBackend::insertNonbulkTriple(const Rdfs::Triple& t, RdfsTable table)
{
    insertSingleTriple(t, table);

    // No triples to infer if predicate is non-transitive or distance is already max
    if (!isTransitive(t.predicate) || t.distance >= m_distance) {
        return;
    }

    QList<Triple> newTriples = findInferredTriples(t, table);

    // Insert triples
    Q_FOREACH(Triple newTriple, newTriples) {
        insertSingleTriple(newTriple, table);
    }
}

void RdfsTripleTBackend::insertSingleTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table)
{
    // Prepare triples for various roles (generic code :) )
    Triple tsop, tpso, tosp;
    tsop.subject = tpso.predicate = tosp.predicate = t.subject;
    tsop.object = tpso.subject = tosp.object = t.predicate;
    tsop.predicate = tpso.object = tosp.subject = t.object;
    tsop.distance = tpso.distance = tosp.distance = t.distance;
    QMap<RdfsRole, Triple> roles;
    QMap<RdfsRole, Triple>::const_iterator i;
    roles.insert(RDFS_SUBJECT, tsop);
    roles.insert(RDFS_PREDICATE, tpso);
    roles.insert(RDFS_OBJECT, tosp);

    for (i = roles.constBegin(); i != roles.constEnd(); ++i) {
        RdfsRole role = i.key();
        Triple tt = i.value();

        QList<TertiaryAtom> secondaryList;
        TripleTPayload data = deserializePayload(payload(tt.subject, role, table));
        if (data.hasData) {
            secondaryList = (data.data.contains(tt.predicate)) ? data.data.value(tt.predicate) : (QList<TertiaryAtom>()); // Secondary list
            insertSorted(TertiaryAtom(tt.object, tt.distance), &secondaryList); // Insert tertiary
            data.data.insert(tt.predicate, secondaryList); // Replace/insert secondary/tertiary
#ifndef NDEBUG
            recomputeStatistics(tt.subject, role, data);
#endif
            storePayload(tt.subject, role, data, table);
        }
        else {
            insertBtreeTriple(tt.subject, role, tt.predicate, tt.object, tt.distance, table);
        }
    }
}

bool RdfsTripleTBackend::isTransitive(const QString& predicate) const
{
    return (predicate == "sp" || predicate == "sc");
}

void RdfsTripleTBackend::storePayload(const QString& primary, const Rdfs::RdfsRole& role, const Rdfs::TripleTPayload& triples, const Rdfs::RdfsTable& table)
{
    RdfsLogger::instance()->debug("Storing primary: " + primary);
    QByteArray data = serializePayload(triples);

    if (data.size() <= m_maxPrimaryPayloadSize) {
        putPayload(primary, role, data, table);
        return;
    }

    RdfsLogger::instance()->debug("Size: " + QString::number(data.size()) + " > " + QString::number(m_maxPrimaryPayloadSize));

    // Store all triples in B-tree
    QMap<QString, QList<TertiaryAtom> >::const_iterator i;
    for (i = triples.data.constBegin(); i != triples.data.constEnd(); ++i) {
        QString secondary = i.key();
        Q_FOREACH(TertiaryAtom tertiary, i.value()) {
            insertBtreeTriple(primary, role, secondary, tertiary.first, tertiary.second, table);
        }
    }

    // Store payload with "no data" marker
    TripleTPayload pointer;
    pointer.hasData = false;
    putPayload(primary, role, serializePayload(pointer), table);
}


void RdfsTripleTBackend::insertSorted(TertiaryAtom item, QList< TertiaryAtom >* list)
{
    if (list->isEmpty()) {
        list->append(item);
        return;
    }

    // Note: we need the intended position of the item, so we cannot reuse binaryFindTriple() here
    int i = 0;
    int j = list->size();
    int mid;
    while (i < j) {
        mid = (i + j) / 2;
        if (list->at(mid).first < item.first) {
            i = mid + 1;
        }
        else {
            j = mid;
        }
    }

    // If item already exists, it will now be at position i
    if (i != list->size() && list->at(i).first == item.first) {
        if (list->at(i).second > item.second) {
            // Replace existing item with new one (with smaller distance)
            list->replace(i, item);
        }

        // Item at i now has minimal distance of the two (existing vs new)
        return;
    }
    // else: item is new (i is position to put it in); i may be end of list, which means append

    list->insert(i, item);
}

void RdfsTripleTBackend::insertBulkTriple(const Rdfs::Triple& t, RdfsTable table)
{
    insertSingleBulkTriple(t, table);

    // Note: if triple distance is max, all inferred triples will have distance higher than max
    if (isTransitive(t.predicate) && t.distance < m_distance) {
        // Find previously materialized triples, which result in inferred triples
        Triple search;
        QList<Triple> newTriples, newTriplesLeft, newTriplesRight;

        // Find triples in current bulk payload (yet-to-be-materialized triples)
        // which are inferred
        QHash<QString, TripleTPayload> *dataset;
        QHash<QString, TripleTPayload>::iterator idata;
        TripleTPayload *data;
        switch(table) {
            default:
            case RDFS_ORIGINAL:
                dataset = &m_bulkPayloadsOriginal;
                break;
            case RDFS_UPDATE:
                dataset = &m_bulkPayloadsUpdate;
                break;
        }

        idata = dataset->find("p:" + t.predicate);
        if (idata != dataset->end()) {
            data = &(idata.value());
            // Note: data is always stored directly in this data set, i.e.
            // no need to use B-trees etc.; those will be found in findInferredTriples()
            // Index order: PSO

            // (?, p, a) + (a, p, b) -> (?, p, b); slightly complicated:
            // Loop over all secondaries (subject) to find out if they form a triple with 'b' in it.
            QMap<QString, QList< TertiaryAtom > >::const_iterator isecondary;
            for (isecondary = data->data.constBegin(); isecondary != data->data.constEnd(); ++isecondary) {
                // Subject: isecondary.key()
                // Do binary search for object 'a' (t.subject).
                QList<TertiaryAtom> tertiaries = isecondary.value();
                QList<TertiaryAtom>::iterator tertiary = binaryFindTertiary(&tertiaries, t.subject);
                if (tertiary != tertiaries.constEnd() && (*tertiary).second + t.distance + 1 <= m_distance) {
                    // Found triple, (*tertiary).first == t.subject
                    Triple newTriple;
                    newTriple.subject = isecondary.key();
                    newTriple.predicate = t.predicate;
                    newTriple.object = t.object;
                    newTriple.distance = (*tertiary).second + t.distance + 1;
                    newTriplesLeft.append(newTriple);
                }
            }

            // (a, p, b) + (b, p, ?) -> (a, p, ?); quite simple since subject is known
            // Just find list of the subject, and add all tertiary triples
            if (data->data.contains(t.object)) {
                QList< TertiaryAtom > tertiaries = data->data.value(t.object);
                Q_FOREACH(TertiaryAtom tertiary, tertiaries) {
                    short unsigned d = t.distance + tertiary.second + 1;
                    if (d <= m_distance) {
                        Triple newTriple;
                        newTriple.subject = t.subject;
                        newTriple.predicate = t.predicate;
                        newTriple.object = tertiary.first;
                        newTriple.distance = tertiary.second + t.distance + 1;
                        newTriplesRight.append(newTriple);
                    }
                }
            }
        }

        // Find materialized left/right triples and merge data
        newTriples = findInferredTriples(t, table, newTriplesLeft, newTriplesRight);

        // All extra data was discovered... move on to insert these triples
        Q_FOREACH(Triple newTriple, newTriples) {
            // This will filter duplicates in insertSorted() (and mergePayloads()
            // in endBulkTriple()).
            insertSingleBulkTriple(newTriple, table);
        }
    }

    if (m_bulkLoadCount >= m_maxBulkLoadSize) {
        // Flush
        endBulkLoad();
        startBulkLoad();
    }
}

void RdfsTripleTBackend::insertSingleBulkTriple(const Rdfs::Triple& t, RdfsTable table)
{
    QHash<QString, TripleTPayload> *dataset;
    switch(table) {
        default:
        case RDFS_ORIGINAL:
            dataset = &m_bulkPayloadsOriginal;
            break;
        case RDFS_UPDATE:
            dataset = &m_bulkPayloadsUpdate;
            break;
    }

    // Add triple to datasets
    QHash<QString, TripleTPayload>::iterator idata;
    TripleTPayload *data;
    QMap<QString, QList< TertiaryAtom > >::iterator subpayload;
    QList<TertiaryAtom> items;

    QList<TripleTPayload *> newPayloads;

    // SOP index
    // Find primary
    idata = dataset->find("s:" + t.subject);
    // Convenience :)
    if (idata != dataset->end()) {
        data = &(idata.value());
    }
    else {
        data = new TripleTPayload();
        data->hasData = true;
        newPayloads.append(data);
    }
    // Find secondary item
    subpayload = data->data.find(t.object);
    if (subpayload != data->data.end()) {
        // Insert tertiary
        insertSorted(TertiaryAtom(t.predicate, t.distance), &(subpayload.value()));
    }
    else {
        // New secondary
        items = QList<TertiaryAtom>();
        items.append(TertiaryAtom(t.predicate, t.distance));
        data->data.insert(t.object, items);
    }
    // New primary
    if (idata == dataset->end()) {
        dataset->insert("s:" + t.subject, *data);
    }

    // PSO
    // Find primary
    idata = dataset->find("p:" + t.predicate);
    // Convenience :)
    if (idata != dataset->end()) {
        data = &(idata.value());
    }
    else {
        data = new TripleTPayload();
        data->hasData = true;
        newPayloads.append(data);
    }
    // Find secondary item
    subpayload = data->data.find(t.subject);
    if (subpayload != data->data.end()) {
        // Insert tertiary
        insertSorted(TertiaryAtom(t.object, t.distance), &(subpayload.value()));
    }
    else {
        // New secondary
        items = QList<TertiaryAtom>();
        items.append(TertiaryAtom(t.object, t.distance));
        data->data.insert(t.subject, items);
    }
    // New primary
    if (idata == dataset->end()) {
        dataset->insert("p:" + t.predicate, *data);
    }

    // OSP
    // Find primary
    idata = dataset->find("o:" + t.object);
    // Convenience :)
    if (idata != dataset->end()) {
        data = &(idata.value());
    }
    else {
        data = new TripleTPayload();
        data->hasData = true;
        newPayloads.append(data);
    }
    // Find secondary item
    subpayload = data->data.find(t.subject);
    if (subpayload != data->data.end()) {
        // Insert tertiary
        insertSorted(TertiaryAtom(t.predicate, t.distance), &(subpayload.value()));
    }
    else {
        // New secondary
        items = QList<TertiaryAtom>();
        items.append(TertiaryAtom(t.predicate, t.distance) );
        data->data.insert(t.subject, items);
    }
    // New primary
    if (idata == dataset->end()) {
        dataset->insert("o:" + t.object, *data);
    }

    // Assuming triple didn't already exist, we have now added 3 items
    m_bulkLoadCount += 3;
    qDeleteAll(newPayloads);
}


QList< Triple > RdfsTripleTBackend::findInferredTriples(const Rdfs::Triple& t, RdfsTable table,
                                                        const QList<Triple> &additionalLeft, const QList<Triple> &additionalRight)
{
    Triple search;
    QList<Triple> inferred, inferredLeft, inferredRight;

    // (?, p, a) + (a, p, b) -> (?, p, b)
    search.subject = QString();
    search.predicate = t.predicate;
    search.object = t.subject;
    inferredLeft = findSubInferredTriples(search, t, table);
    inferredLeft.append(additionalLeft);

    // (a, p, b) + (b, p, ?) -> (a, p, ?)
    search.subject = t.object;
    search.predicate = t.predicate;
    search.object = QString();
    inferredRight = findSubInferredTriples(search, t, table);
    inferredRight.append(additionalRight);

    // Merge stuff: (?1, p, a) + (a, p, b) + (b, p, ?2) -> (?1. p. ?2)
    // Use already found triples to get this data
    Q_FOREACH(Triple l, inferredLeft) {
        Q_FOREACH(Triple r, inferredRight) {
            // Note: base distance is 0, so no need to correct distance for center triple
            unsigned char d = l.distance + r.distance;
            if (d <= m_distance) {
                Triple newTriple;
                newTriple.subject = l.subject;
                newTriple.predicate = l.predicate;
                newTriple.object = r.object;
                newTriple.distance = d;
                inferred.append(newTriple);
            }
        }
    }

    inferred.append(inferredLeft);
    inferred.append(inferredRight);

    return inferred;
}

QList< Triple > RdfsTripleTBackend::findSubInferredTriples(const Rdfs::Triple& search, const Rdfs::Triple& original, RdfsTable table)
{
    QList<Triple> resultTriples;

    // Search triples
    RdfsCursor *c = cursor(search, table);
    if (c->open()) {
        Triple result, newTriple;
        while (!c->eof()) {
            result = c->current();
            if (result.distance + original.distance + 1 <= m_distance) {
                newTriple.subject = (search.subject.isEmpty()) ? result.subject : original.subject;
                newTriple.predicate = original.predicate;
                newTriple.object = (search.object.isEmpty()) ? result.object : original.object;
                newTriple.distance = result.distance + original.distance + 1;
                // No need to recurse: all triples should be materialized in the first cursor
                resultTriples.append(newTriple);
            }
            c->next();
        }
        c->close();
    }
    delete c;

    return resultTriples;
}


quint32 Rdfs::RdfsTripleTBackend::payloadSize(TripleTPayload payload)
{
    quint32 size = 3*sizeof(quint32);
    QMap<QString, QList<TertiaryAtom> >::const_iterator i;
    for (i = payload.data.constBegin(); i != payload.data.constEnd(); ++i) {
        size += i.key().size() + 2*sizeof(quint32);
        Q_FOREACH(TertiaryAtom atom, i.value()) {
            size += atom.first.size() + sizeof(quint32) + sizeof(unsigned char);
        }
    }
    return size;
}


void RdfsTripleTBackend::startBulkLoad()
{
    m_inBulkLoad = true;
}

void RdfsTripleTBackend::endBulkLoad()
{
    QHash<QString, TripleTPayload>::const_iterator i;
    quint32 count = 0;
    for (int t = 0; t < 2; ++t) {
        RdfsTable table;
        QHash<QString, TripleTPayload> container;
        switch(t) {
            case 0:
                table = RDFS_ORIGINAL;
                container = m_bulkPayloadsOriginal;
                break;
            case 1:
                table = RDFS_UPDATE;
                container = m_bulkPayloadsUpdate;
                break;
        }

        for (i = container.constBegin(); i != container.constEnd(); ++i) {
            QString key = i.key();
            RdfsRole role;
            if (key.startsWith("s:")) {
                role = RDFS_SUBJECT;
            }
            else if (key.startsWith("p:")) {
                role = RDFS_PREDICATE;
            }
            else {
                role = RDFS_OBJECT;
            }
            key = key.remove(0, 2);

            TripleTPayload original = deserializePayload(payload(key, role, table));

            if (original.hasData) {
                RdfsLogger::instance()->debug("Merging");
                TripleTPayload data = mergePayloads(original, i.value());
                storePayload(key, role, data, table);
#ifndef NDEBUG
                recomputeStatistics(key, role, data);
#endif
            }
            else {
                RdfsLogger::instance()->debug("Original: no data - using BTree");
                // Original in B-tree, so add all new triples to B-tree
                QMap<QString, QList<TertiaryAtom> >::const_iterator secondary;
                for (secondary = i.value().data.constBegin(); secondary != i.value().data.constEnd(); ++secondary) {
                    Q_FOREACH(TertiaryAtom tertiary, secondary.value()) {
                        insertBtreeTriple(key, role, secondary.key(), tertiary.first, tertiary.second, table);
                    }
                }
            }

            ++count;
        }
    }

    m_bulkLoadCount = 0;
    m_bulkPayloadsOriginal.clear();
    m_bulkPayloadsUpdate.clear();
    m_inBulkLoad = false;

    RdfsLogger::instance()->debug(QString("Atoms bulk inserted: %1").arg(count));
}

TripleTPayload RdfsTripleTBackend::mergePayloads(const TripleTPayload& a, const TripleTPayload& b) const
{
    TripleTPayload result;
    result.hasData = true; // Assume a and b have data (or are empty)

    QMap<QString, QList<TertiaryAtom> >::const_iterator ia, ib;
    for (ia = a.data.constBegin(), ib = b.data.constBegin(); ib != b.data.constEnd(); ++ib) {
        // Find position in a, linear search
        while (ia != a.data.constEnd() && ia.key() < ib.key()) {
            result.data.insert(ia.key(), ia.value());
            ++ia;
        }

        if (ia == a.data.constEnd() || ia.key() != ib.key()) {
            // Either no more items in A, or next item should be put after current B
            result.data.insert(ib.key(), ib.value());
        }
        else {
            // Equal key: merge values
            result.data.insert(ia.key(), mergeLists(ia.value(), ib.value()));
            ++ia;
        }
    }
    // B is exhausted, now append any remaining values in A
    while (ia != a.data.constEnd()) {
        result.data.insert(ia.key(), ia.value());
        ++ia;
    }

    return result;
}


QList< TertiaryAtom > RdfsTripleTBackend::mergeLists(const QList< TertiaryAtom > &a, const QList< TertiaryAtom > &b) const
{
    QList<TertiaryAtom> result;

    QList<TertiaryAtom>::const_iterator ia, ib;
    for (ia = a.constBegin(), ib = b.constBegin(); ib != b.constEnd(); ++ib) {
        while (ia != a.constEnd() && (*ia).first < (*ib).first) {
            result.append(*ia);
            ++ia;
        }
        // (*ia).first >= (*ib.first) -- keep the one w/ smallest distance if equal, otherwise append b
        if (ia != a.constEnd() && (*ia).first == (*ib).first && (*ia).second < (*ib).second) {
            result.append(*ia);
        }
        else { // ia.first > ib.first, or ia.second >= ib.second, or ia at end
            result.append(*ib);
        }

        if (ia != a.constEnd() && (*ia).first == (*ib).first) { // Duplicate tertiary - skip ia
            ++ia;
        }
    }

    // Add remaining items
    while (ia != a.constEnd()) {
        result.append(*ia);
        ++ia;
    }

    return result;
}


bool RdfsTripleTBackend::hasTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table)
{
    bool result = false;;

    RdfsCursor *c = cursor(t, table);
    if (c->open()) {
        result = true;
        c->close();
    }
    delete c;

    return result;
}

Rdfs::RdfsCursor* RdfsTripleTBackend::cursor(Rdfs::Triple q, Rdfs::RdfsTable table)
{
    return new RdfsTripleTCursor(this, q, table);
}

void RdfsTripleTBackend::insertRule(Rule r)
{
    // TODO STUB
}

void RdfsTripleTBackend::removeRule(Rule r)
{
    // TODO STUB
}

QList< Rule > RdfsTripleTBackend::rules()
{
    // TODO STUB
    return QList< Rule >();
}

#ifndef NDEBUG
void RdfsTripleTBackend::loadStatistics()
{
    Triple t;
    t.subject = "rdfsupdate:ttstats";
    t.predicate = "rdfsupdate:tcount";
    QList<Triple> statlist = findTriples(t, RDFS_ORIGINAL);

    if (!statlist.isEmpty()) {
        bool ok;
        m_tripleCount = statlist.at(0).object.toULong(&ok);
    }
    else {
        m_tripleCount = 0;
    }

    t.predicate = "rdfsupdate:datasize";
    statlist = findTriples(t, RDFS_ORIGINAL);
    if (!statlist.isEmpty()) {
        bool ok;
        m_dataSize = statlist.at(0).object.toULong(&ok);
    }
    else {
        m_dataSize = 0;
    }

    m_loadedStatistics = true;
}

void RdfsTripleTBackend::saveStatistics()
{
    if (m_savingStatistics) {
        return;
    }
    if (!m_loadedStatistics) {
        loadStatistics();
    }

    m_savingStatistics = true;

    Triple t;
    QList<Triple> remove;

    // Remove any old statistical data
    t.subject = "rdfsupdate:ttstats";
    t.predicate = "rdfsupdate:tcount";
    t.object = QString();
    remove = findTriples(t, RDFS_ORIGINAL);
    Q_FOREACH(t, remove) {
        removeTriple(t);
    }

    t.subject = "rdfsupdate:ttstats";
    t.predicate = "rdfsupdate:datasize";
    t.object = QString();
    remove = findTriples(t, RDFS_ORIGINAL);
    Q_FOREACH(t, remove) {
        removeTriple(t);
    }

    // Insert new data
    t.subject = "rdfsupdate:ttstats";
    t.predicate = "rdfsupdate:tcount";
    t.object = QString::number(m_tripleCount);
    insertTriple(t);

    t.subject = "rdfsupdate:ttstats";
    t.predicate = "rdfsupdate:datasize";
    t.object = QString::number(m_dataSize);
    insertTriple(t);

    m_savingStatistics = false;
}

void RdfsTripleTBackend::recomputeStatistics(const QString& atom, const RdfsRole &role, const Rdfs::TripleTPayload &payload)
{
    if (m_savingStatistics) {
        return; // Do not count statistics-triples
    }

    if (!m_loadedStatistics) {
        loadStatistics();
    }

    QByteArray oldData = this->payload(atom, role);

    if (!oldData.isEmpty()) {
        TripleTPayload p = deserializePayload(oldData);
        m_tripleCount -= payloadTripleCount(p);
        m_dataSize -= oldData.size();
    }

    m_tripleCount += payloadTripleCount(payload);
    m_dataSize += serializePayload(payload).size();
}

uint RdfsTripleTBackend::payloadTripleCount(const Rdfs::TripleTPayload& payload)
{
    uint n = 0;
    QMap<QString, QList<TertiaryAtom> >::const_iterator i;
    for (i = payload.data.constBegin(); i != payload.data.constEnd(); ++i) {
        n += i.value().size();
    }
    return n;
}
#endif

RdfsTripleTBackend::~RdfsTripleTBackend()
{

}

