/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "rdfstripletcursor.h"
#include "rdfslogger.h"

using namespace Rdfs;

RdfsTripleTCursor::RdfsTripleTCursor(Rdfs::RdfsTripleTBackend* backend, const Rdfs::Triple& query, Rdfs::RdfsTable table)
    : RdfsCursor(backend, query, table)
{
    m_ttbackend = backend;
    m_btreeCursor = 0;
    m_primary = m_primaryList.constEnd();
    m_secondary = m_payload.constEnd();
    m_tertiary = m_tertiaryList.constEnd();
}

RdfsTripleTCursor::~RdfsTripleTCursor()
{
    close();
}

bool RdfsTripleTCursor::open()
{
    Triple t = queryTriple();
    RdfsTable table = queryTable();

    m_primaryList = QStringList();

    if (!t.subject.isEmpty()) {
        m_primaryList << t.subject;
        m_order = RDFS_SUBJECT;
        RdfsLogger::instance()->debug("SOP");
        m_orderedQuery.subject = t.subject;
        m_orderedQuery.predicate = t.object;
        m_orderedQuery.object = t.predicate;
    }
    else if (!t.object.isEmpty()) {
        m_primaryList << t.object;
        m_order = RDFS_OBJECT;
        RdfsLogger::instance()->debug("OSP");
        m_orderedQuery.subject = t.object;
        m_orderedQuery.predicate = t.subject;
        m_orderedQuery.object = t.predicate;
    }
    else if (!t.predicate.isEmpty()) {
        m_primaryList << t.predicate;
        m_order = RDFS_PREDICATE;
        RdfsLogger::instance()->debug("PSO");
        m_orderedQuery.subject = t.predicate;
        m_orderedQuery.predicate = t.subject;
        m_orderedQuery.object = t.object;
    }
    else { // (t.subject.isEmpty() && t.predicate.isEmpty() && t.object.isEmpty())
        // Special case: no indexing possible if all three atoms are empty
        m_primaryList = m_ttbackend->atoms(table);
        m_order = RDFS_SUBJECT;
        RdfsLogger::instance()->debug("SOP EMPTY");
        m_orderedQuery.subject = t.subject;
        m_orderedQuery.predicate = t.object;
        m_orderedQuery.object = t.predicate;
    }

    // Initialize primary (which will init secondary + tertiary)
    m_primary = m_primaryList.constBegin();
    nextTertiary(true);

    // If m_primary == primaryList.constEnd(), there are no matches on primary
    return (!eof());
}

void RdfsTripleTCursor::nextPrimary(bool init)
{
    // Reset secondary/tertiary
    m_payload.clear();
    m_secondary = m_payload.constEnd();
    m_tertiaryList.clear();
    m_tertiary = m_tertiaryList.constEnd();

    if (m_btreeCursor) {
        m_btreeCursor->close();
        delete m_btreeCursor;
        m_btreeCursor = 0;
    }

    bool needNext;
    do {
        needNext = false; // Assume we'll find a match by default
        // Advance to next item
        if (!init) {
            m_primary++;
        }
        init = false;

        if (m_primary == m_primaryList.constEnd()) {
            return;
        }

        QString atom = *m_primary;
        RdfsLogger::instance()->debug("Primary atom: " + atom);

        // Advance until valid match
        if (!m_orderedQuery.subject.isEmpty()) {
            while (m_orderedQuery.subject != atom) {
                m_primary++;
                if (m_primary == m_primaryList.constEnd()) {
                    return; // Unable to continue
                }
                atom = *m_primary;
            }
        }

        TripleTPayload payload = m_ttbackend->deserializePayload(m_ttbackend->payload(atom, m_order, queryTable()));

        // Determine whether to use a BTree or "normal" cursor to proceed
        if (payload.hasData) {
            if (m_orderedQuery.predicate.isEmpty()) {
                m_payload = payload.data;
                m_secondary = m_payload.constBegin();
                needNext = m_payload.isEmpty();
            }
            else if (payload.data.contains(m_orderedQuery.predicate)) {
                m_payload.insert(m_orderedQuery.predicate, payload.data.value(m_orderedQuery.predicate));
                m_secondary = m_payload.constBegin();
            }
            else {
                needNext = true;
            }
        }
        else {
            m_btreeCursor = m_ttbackend->btreeCursor(atom, m_order, m_orderedQuery.predicate, m_orderedQuery.object, queryTable());
            if (!m_btreeCursor->open()) {
                // No valid items -> continue to next primary again
                needNext = true;
            }
        }
    } while (needNext);
}

void RdfsTripleTCursor::nextSecondary(bool init)
{
    // On init, first get primary
    if (init) {
        nextPrimary(true);
    }

    // Reset tertiary
    m_tertiaryList.clear();
    m_tertiary = m_tertiaryList.constEnd();

    if (eof()) {
        return;
    }

    bool needNext, advance;
    advance = !init;
    do {
        needNext = false; // Assume we'll get a hit by default
        if (advance) {
            if (!m_btreeCursor) {
                do {
                    m_secondary++;
                } while (m_secondary != m_payload.constEnd() &&
                         !m_orderedQuery.predicate.isEmpty() && // Empty secondary key is always a hit
                         m_secondary.key() != m_orderedQuery.predicate);
                // while: make sure we meet a hit, or the end of the primary data

                // We need to advance the primary if we hit the end
                if (m_secondary == m_payload.constEnd()) {
                    nextPrimary(); // Will position secondary correctly, possibly with btree
                }
            }
            else {
                // We're called from nextTertiary and hit the end of the BTreeCursor.
                // I.e., we need the next primary, which may result in a "normal"
                // secondary cursor
                nextPrimary(); // Will position secondary correctly
            }

            if (eof()) { // We have have moved primary beyond end - just return
                return;
            }
        }
        // Always advance cursor next round
        advance = true;

        // Secondary is now at position - prepare tertiary cursor
        if (!m_btreeCursor) {
            if (!m_orderedQuery.object.isEmpty()) {
                // Binary find to tertiary - if found, prepare list, otherwise
                // we need the next secondary
                QList<TertiaryAtom> items = m_secondary.value();
                QList<TertiaryAtom>::iterator iitem;
                iitem = RdfsTripleTBackend::binaryFindTertiary(&items, m_orderedQuery.object);
                if (iitem != items.end()) {
                    // Found match, prepare tertiary (new list, single item, add in distance from binary find result)
                    m_tertiaryList << TertiaryAtom((*iitem).first, (*iitem).second);
                    m_tertiary = m_tertiaryList.constBegin();
                }
                else {
                    // Reset, get next secondary
                    m_tertiary = m_tertiaryList.constEnd();
                    needNext = true;
                }
            }
            else {
                // Go over all tertiaries
                // Make sure list isn't empty, just in case
                m_tertiaryList = m_secondary.value();
                m_tertiary = m_tertiaryList.constBegin();

                if (m_tertiaryList.isEmpty()) {
                    // constBegin == constEnd -> no need to reset m_tertiary
                    needNext = true;
                }
            }
        }
        // If btreecursor: tertiary will also be in place

    } while (needNext);
}

void RdfsTripleTCursor::nextTertiary(bool init)
{
    // On init, first get secondary
    if (init) {
        nextSecondary(true);
    }
    if (eof()) {
        return;
    }

    if (!init && !m_btreeCursor) {
        // Advance tertiary
        m_tertiary++;

        if (m_tertiary == m_tertiaryList.constEnd()) {
            nextSecondary(); // Get next secondary item, this will also position the
                             // tertiary cursor correctly
                             // Note that this _may_ cause a btreeCursor to be set!
        }
    }
    else if (!init && m_btreeCursor) {
        // Advance btree cursor
        m_btreeCursor->next();

        if (m_btreeCursor->eof()) {
            nextSecondary(); // May delete btree cursor! Note that we cannot
                             // immediately call nextPrimary(), as that will not
                             // position the secondary cursor if the next primary
                             // has no btree cursor.
        }
    }
    // Tertiary is now in position
}

bool RdfsTripleTCursor::next()
{
    nextTertiary();
    return (!eof());
}

bool RdfsTripleTCursor::eof() const
{
    // If primary moves beyond end of primary list, we have no triples anymore
    return (m_primary == m_primaryList.constEnd());
}

void RdfsTripleTCursor::close()
{
    // If there's a B-tree cursor: clean it up
    if (m_btreeCursor) {
        m_btreeCursor->close();
        delete m_btreeCursor;
        m_btreeCursor = 0;
    }

    // Clear the rest
    m_primaryList.clear();
    m_primary = m_primaryList.constEnd();

    m_payload.clear();
    m_secondary = m_payload.constEnd();
    m_tertiaryList.clear();
    m_tertiary = m_tertiaryList.constEnd();
}

Rdfs::Triple RdfsTripleTCursor::current() const
{
    Triple ordered, result;

    if (eof()) {
        return result; // Empty triple
    }

    if (m_btreeCursor) {
        ordered = m_btreeCursor->current();
    }
    else {
        ordered.subject = *m_primary;
        ordered.predicate = m_secondary.key();
        ordered.object = (*m_tertiary).first;
        ordered.distance = (*m_tertiary).second;
    }

    switch(m_order) {
        default:
        case RDFS_SUBJECT:
            result.subject = ordered.subject;
            result.predicate = ordered.object;
            result.object = ordered.predicate;
            break;
        case RDFS_PREDICATE:
            result.subject = ordered.predicate;
            result.predicate = ordered.subject;
            result.object = ordered.object;
            break;
        case RDFS_OBJECT:
            result.subject = ordered.predicate;
            result.predicate = ordered.object;
            result.object = ordered.subject;
            break;
    }
    result.distance = ordered.distance;
    return result;
}


