/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "rdfsupdate.h"

#include <QCoreApplication>
#include <QTextStream>
#include <QFile>
#include <QSocketNotifier>
#include <QString>
#include <QStringList>
#include <QFileInfo>
#include <QUrl>
#include <QSqlError>

#include <iostream>
#include <stdio.h>
#include <boost/tokenizer.hpp>
#include <string>

#include "rdfssqlite.h"
#include "rdfslogger.h"
#include "rdfscommand.h"

#include "commands/rdfshelp.h"
#include "commands/rdfssqlconnect.h"
#include "commands/rdfssqldisconnect.h"
#include "commands/rdfslistoptions.h"
#include "commands/rdfssetoption.h"
#include "commands/rdfsinsert.h"
#include "commands/rdfsremove.h"
#include "commands/rdfslist.h"
#include "commands/rdfschecktriple.h"
#include "commands/rdfsclosurecheck.h"
#include "commands/rdfsloadfile.h"
#include "commands/rdfsexport.h"
#include "commands/rdfstruncate.h"
#include "commands/rdfsaddrule.h"
#include "commands/rdfslistrules.h"
#include "commands/rdfsremoverule.h"
#include "commands/rdfsunchase.h"
#include "commands/rdfsstats.h"
#include "commands/rdfschainstats.h"
#include "commands/rdfsdotexport.h"
#include "commands/rdfsgraphgen.h"
#include "commands/rdfskillcycles.h"
#include "commands/instancegenerate.h"
#include "commands/rdfsontologylinks.h"
#include "commands/rdfsontologydot.h"
#include "commands/findqueries.h"
#include "commands/srand.h"
#include "commands/rdfsquit.h"

#ifdef VALGRIND
    #include "commands/callgrind.h"
#endif

using namespace Rdfs;

RdfsUpdate::RdfsUpdate(QObject *parent)
    : QObject(parent), m_busy(false), m_store(0)
{
    initCommands();

    m_in = new QFile();
    m_in->open(stdin, QIODevice::ReadOnly);

    m_inNotifier = new QSocketNotifier(0, QSocketNotifier::Read);

    connect(m_inNotifier, SIGNAL(activated(int)), SLOT(readCommand()));

    // Show license info
    RdfsLogger::instance()->log(tr("  RdfsUpdate"));
    RdfsLogger::instance()->log(tr("  Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl"));
    RdfsLogger::instance()->log(tr("  Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>"));
    RdfsLogger::instance()->log(tr("  This program comes with ABSOLUTELY NO WARRANTY."));
    RdfsLogger::instance()->log(tr("  This is free software, and you are welcome to redistribute it"));
    RdfsLogger::instance()->log(tr("  under certain conditions. See license.txt for more details."));
    RdfsLogger::instance()->log(QString());

    showPrompt();
}


void RdfsUpdate::initCommands() {
    // Basics
    registerCommand(new RdfsHelp(this));
    registerCommand(new RdfsSqlConnect(this));
    registerCommand(new RdfsSqlDisconnect(this));
    registerCommand(new RdfsListOptions(this));
    registerCommand(new RdfsSetOption(this));
    registerCommand(new Srand(this));

    // Graph management
    registerCommand(new RdfsCheckTriple(this));
    registerCommand(new RdfsLoadFile(this));
    registerCommand(new RdfsExport(this));
    registerCommand(new RdfsInsert(this));
    registerCommand(new RdfsRemove(this));
    registerCommand(new RdfsTruncate(this));
    registerCommand(new RdfsList(this));
    registerCommand(new RdfsChainStats(this));

    // Rule management
    registerCommand(new RdfsListRules(this));
    registerCommand(new RdfsAddRule(this));
    registerCommand(new RdfsRemoveRule(this));

    // The two main algorithms
    registerCommand(new RdfsClosureCheck(this));
    registerCommand(new RdfsUnchase(this));

    // Statistics
    registerCommand(new RdfsStats(this));
    registerCommand(new RdfsDotExport(this));

    // Data generation
    registerCommand(new RdfsGraphGen(this));
    registerCommand(new RdfsKillCycles(this));
    registerCommand(new RdfsOntologyLinks(this));
    registerCommand(new RdfsOntologyDot(this));
    registerCommand(new InstanceGenerate(this));
    registerCommand(new FindQueries(this));

    // Valgrind
#ifdef VALGRIND
    registerCommand(new Callgrind(this));
#endif

    // You don't really want this!
    registerCommand(new RdfsQuit(this));
}


void RdfsUpdate::registerCommand(RdfsCommand* command)
{
    if (m_commands.contains(command->name())) {
        RdfsLogger::instance()->debug(tr("Warning! Command name clash: %1").arg(command->name()));
    }
    m_commands.insert(command->name(), command);

    QString alias;
    Q_FOREACH(alias, command->aliases()) {
        if (m_commands.contains(alias)) {
            RdfsLogger::instance()->debug(tr("Warning! Command alias clash: %1").arg(alias));
        }
        m_commands.insert(alias, command);
    }

    m_commandList.append(command);
}

QList<RdfsCommand*> RdfsUpdate::commands() {
    return m_commandList;
}


RdfsBackend* RdfsUpdate::store() {
    return m_store;
}


void RdfsUpdate::setStore(RdfsBackend* store) {
    m_store = store;
}


void RdfsUpdate::readCommand()
{
    if (m_busy) {
        return;
    }

    // Temporarily disable notifier
    m_inNotifier->setEnabled(false);
    m_busy = true;

    // Read line
    QString line(m_in->readLine());

    // End of file: exit app (line will be empty)
    if (m_in->atEnd()) {
        QCoreApplication::exit();
        m_busy = false;
        m_inNotifier->setEnabled(true);
        return;
    }

    line = line.trimmed();

    // Reprint the echo to the output if necessary
    RdfsLogger::instance()->reprint(line);

    // Ignore empty lines, comments - reshow prompt
    if (line.isEmpty() || line.startsWith("#")) {
        m_busy = false;
        m_inNotifier->setEnabled(true);
        showPrompt();
        return;
    }

    // Split command and arguments
    QString command;
    QStringList arguments;

    // Go through boost tokenizer - use space as field separator
    std::string s = line.toStdString();
    boost::escaped_list_separator<char> el('\\', ' ', '\"');
    boost::tokenizer<boost::escaped_list_separator<char> > tok(s, el);
    for(boost::tokenizer<boost::escaped_list_separator<char> >::iterator beg=tok.begin(); beg!=tok.end();++beg) {
        // Remove
        QString arg = QString::fromStdString(*beg).trimmed();
        if (!arg.isEmpty()) {
            arguments << arg;
        }
    }
    command = arguments.takeFirst();

    // Dispatch command
    execute(command, arguments);

    m_busy = false;
    m_inNotifier->setEnabled(true);
    showPrompt();
}

bool RdfsUpdate::execute(const QString &command, const QStringList &arguments) {
    // Check if valid command
    if (!m_commands.contains(command)) {
        RdfsLogger::instance()->log(tr("Unknown command: %1").arg(command));
        RdfsLogger::instance()->log(tr("Enter 'help' for a list of known commands."));
        return false;
    }

    // Load command
    RdfsCommand *cmd = m_commands.value(command);
    if (cmd->requireConnection() && !isConnected()) {
        RdfsLogger::instance()->log(tr("Open a database connection first."));
        return false;
    }

    // Check arguments
    int minArgCount = 0;
    int maxArgCount = 0;
    CommandParameter p;
    Q_FOREACH(p, cmd->arguments()) {
        ++maxArgCount;
        if (p.required) {
            ++minArgCount;
        }
    }

    // Check count
    if (minArgCount > arguments.count() || maxArgCount < arguments.count()) {
        RdfsLogger::instance()->log(tr("Wrong number of arguments."));
        this->usage(cmd);
        return false;
    }

    // Validate arguments
    QString error = cmd->checkArguments(arguments);
    if (!error.isEmpty()) {
        RdfsLogger::instance()->log(tr("Argument error: %1").arg(error));
        this->usage(cmd);
        return false;
    }

    // All ok: execute command
    return cmd->exec(arguments);
}

RdfsCommand* RdfsUpdate::command(QString name) {
    if (!m_commands.contains(name)) {
        return 0;
    }
    else {
        return m_commands.value(name);
    }
}

void RdfsUpdate::showPrompt() const {
    RdfsLogger::instance()->prompt("rdfsupdate > ");
}


bool RdfsUpdate::isConnected() {
    if (m_store == 0) {
        return false;
    }
    return m_store->isOpen();
}

void RdfsUpdate::disconnectStore() {
    if (m_store != 0) {
        delete m_store; // Will/should disconnect in destructor
        m_store = 0;
        RdfsLogger::instance()->log(tr("Closed database connection"));
    }
}

void RdfsUpdate::usage(RdfsCommand* command) {
    // Initialize argument help
    QStringList argHelpList;
    QString argumentsList;
    QString argHelp;
    QString argExample;
    CommandParameter argument;
    Q_FOREACH(argument, command->arguments()) {
        if (!argument.required) {
            argumentsList += " [" + argument.name + "]";
        }
        else {
            argumentsList += " " + argument.name;
        }

        argHelp = QString("    %1: %2\n").arg(argument.name, 12).arg(argument.help);
        Q_FOREACH(argExample, argument.examples) {
            argHelp += QString("                  " + tr("e.g. %1") + "\n").arg(argExample);
        }
        argHelpList << argHelp;
    }

    // Basic "command: info" string
    RdfsLogger::instance()->log(QString("%1: %2").arg(command->name()).arg(command->description()));
    // Usage line
    RdfsLogger::instance()->log(tr("Usage: %1 %2").arg(command->name()).arg(argumentsList));
    if (command->requireConnection()) {
        RdfsLogger::instance()->log(tr("This command needs a database connection."));
    }

    if (argHelpList.size() > 0) {
        RdfsLogger::instance()->log(tr("Arguments:"));
        RdfsLogger::instance()->log(argHelpList.join("\n"));
    }
    else {
        RdfsLogger::instance()->log(tr("This command does not accept any arguments."));
    }
}

RdfsUpdate::~RdfsUpdate() {
    // Log empty line so output ends with empty line (and shell prompt is on new line)
    RdfsLogger::instance()->log(QString());

    disconnectStore();
    m_in->close();
    delete m_in;
    delete m_inNotifier;
}

#include "rdfsupdate.moc"
