/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef rdfsupdate_H
#define rdfsupdate_H

#include <QObject>
#include <QMap>
#include <QList>

// Required for RdfsTable enum
#include "rdfsbackend.h"

class QFile;
class QSocketNotifier;

namespace Rdfs {

class RdfsSql;
class RdfsCommand;
struct Triple;

class RdfsUpdate : public QObject
{
Q_OBJECT
private:
    /** Stdin file handle */
    QFile *m_in;
    /** Notifier for new input */
    QSocketNotifier *m_inNotifier;
    /** Indicator if we're busy (deal with only one line at a time) */
    bool m_busy;

    /** Sql Database API */
    RdfsBackend *m_store;

    /** Map of commands (command name to command implementation) */
    QMap<QString, RdfsCommand *> m_commands;

    /** Plain list of commands, used for inspection. */
    QList<RdfsCommand *> m_commandList;

    /**
     * Initialize the list of commands.
     */
    void initCommands();

    /**
     * Execute a command with the given arguments.
     * @param command The command to execute
     * @param arguments List of arguments to pass on
     */
    bool execute(const QString &command, const QStringList &arguments);

private slots:
    /**
     * Read and execute a command from the standard input.
     */
    void readCommand();

    /**
     * Registers a new command that can be executed.
     * @param command The command to make available to the user
     */
    void registerCommand(RdfsCommand* command);

    /**
     * Show a prompt at stdout
     */
    inline void showPrompt() const;

public:
    /**
     * Main constructor
     */
    explicit RdfsUpdate(QObject *parent = 0);

    /**
     * Destructor
     */
    virtual ~RdfsUpdate();

    /**
     * Get a list of available commands
     */
    QList<RdfsCommand *> commands();

    /**
     * Get the current database (if any). May be NULL if no connection was made.
     */
    RdfsBackend *store();

    /** Check if a database connection is open */
    bool isConnected();

    /**
     * Get a command by name or alias
     * @param name The name (or alias) of the command to retrieve
     */
    RdfsCommand *command(QString name);

public slots:
    /** Disconnect the SQL database, if open */
    void disconnectStore();

    /**
     * Set a storage backend as current connection.
     * @param store The storage backend connection to set.
     */
    void setStore(RdfsBackend *store);

    /**
     * Show the usage of a specific command
     */
    void usage(RdfsCommand* command);
};
};

#endif // rdfsupdate_H
